import pytest
import xarray as xr

from ravenpy.config.commands import (
    LandUseClassesCommand,
    SBGroupPropertyMultiplierCommand,
    SoilClassesCommand,
    SoilProfilesCommand,
    VegetationClassesCommand,
)
from ravenpy.extractors.routing_product import (
    RoutingProductGridWeightExtractor,
    RoutingProductShapefileExtractor,
)
from ravenpy.models import Raven
from ravenpy.utilities.testdata import get_local_testdata


class TestRouting:
    def test_lievre_tutorial(self):
        """
        This test reproduces the Lievre tutorial setup:

        http://raven.uwaterloo.ca/files/RavenTutorial6.zip

        """

        ###############
        # Input files #
        ###############

        routing_product_shp_path = get_local_testdata(
            "raven-routing-sample/finalcat_hru_info.zip"
        )

        vic_streaminputs_nc_path = get_local_testdata(
            "raven-routing-sample/VIC_streaminputs.nc"
        )
        vic_temperatures_nc_path = get_local_testdata(
            "raven-routing-sample/VIC_temperatures.nc"
        )

        observation_data_nc_path = get_local_testdata(
            "raven-routing-sample/WSC02LE024.nc"
        )

        #########
        # Model #
        #########

        model = Raven()

        model.config.identifier = "raven-lievre-routing"

        #######
        # RVI #
        #######

        model.config.rvi.tmpl = """
        :CatchmentRoute        ROUTE_DUMP                    # Catchment routing method, used to convey water from the catchment tributaries and rivulets to the subbasin outlets. DEFAULT ROUTE_DUMP, which instantly ‘dumps’ all water in the subbasin stream reach.
        :Routing               ROUTE_DIFFUSIVE_WAVE          # Channel routing method which is used to transport water from upstream to downstream within the main subbasin channels. DEFAULT ROUTE_DIFFUSIVE_WAVE, which analytically solves the diffusive wave equation along the reach using a constant reference celerity.
        :PrecipIceptFract      PRECIP_ICEPT_NONE             # Estimation of the precipitation interception fraction. In this routing model, stream input(s) are "pretending" to be precipitation going into Raven, thus using DEFAULT PRECIP_ICEPT_NONE to indicate no interception processes are adopted.
        :PotentialMeltMethod   POTMELT_NONE                  # Estimation of the potential snow melt. In this routing model, snow melt processes are not relevant, thus using DEFAULT POTMELT_NONE method.
        :SoilModel             SOIL_ONE_LAYER                # In this routing model, use DEFAULT SOIL_ONE_LAYER to define single soil layer structure.

        :HydrologicProcesses
          :Precipitation     PRECIP_RAVEN             ATMOS_PRECIP     PONDED_WATER          # Moves stream input(s) from ATMOS_PRECIP to PONDED_WATER storage (waiting for runoff). Use DEFAULT PRECIP_RAVEN method.
          :Flush             RAVEN_DEFAULT            PONDED_WATER     SURFACE_WATER         # Moves water from PONDED_WATER to SURFACE_WATER (routed to outlet). Use DEFAULT RAVEN_DEFAULT method.
        :EndHydrologicProcesses
        """

        streaminputs = xr.open_dataset(vic_streaminputs_nc_path)

        start = streaminputs.indexes["time"][0]
        end = streaminputs.indexes["time"][-4]  # to match the tutorial end date

        model.config.rvi.start_date = start.to_pydatetime()
        model.config.rvi.end_date = end.to_pydatetime()

        # Raven will use 24h even though the NC inputs are 6h
        model.config.rvi.time_step = "24:00:00"

        model.config.rvi.evaluation_metrics = [
            "NASH_SUTCLIFFE",
            "PCT_BIAS",
            "KLING_GUPTA",
        ]

        #######
        # RVH #
        #######

        rvh_extractor = RoutingProductShapefileExtractor(
            routing_product_shp_path, hru_aspect_convention="ArcGIS"
        )
        rvh_config = rvh_extractor.extract()
        channel_profiles = rvh_config.pop("channel_profiles")

        gauge = [sb for sb in rvh_config["subbasins"] if sb.gauged]
        assert len(gauge) == 1
        gauge = gauge[0]

        rvh_config.update(
            {
                "land_subbasin_property_multiplier": SBGroupPropertyMultiplierCommand(
                    "Land", "MANNINGS_N", 1.0
                ),
                "lake_subbasin_property_multiplier": SBGroupPropertyMultiplierCommand(
                    "Lakes", "RESERVOIR_CREST_WIDTH", 1.0
                ),
            }
        )

        for k, v in rvh_config.items():
            model.config.rvh.update(k, v)

        #######
        # RVP #
        #######

        # The labels used for the following commands ("Lake_Soil_Lake_HRU", "Soil_Land_HRU", etc)
        # must correspond to the values of certain fields of the Routing Product:
        # LAND_USE_C, VEG_C, SOIL_PROF

        model.config.rvp.tmpl = """
        {soil_classes}

        {soil_profiles}

        {vegetation_classes}

        {land_use_classes}

        {avg_annual_runoff}

        {channel_profiles}
        """

        model.config.rvp.avg_annual_runoff = 594
        model.config.rvp.soil_classes = [SoilClassesCommand.Record("AQUIFER")]
        model.config.rvp.soil_profiles = [
            SoilProfilesCommand.Record("Lake_Soil_Lake_HRU", ("AQUIFER",), (5,)),
            SoilProfilesCommand.Record("Soil_Land_HRU", ("AQUIFER",), (5,)),
        ]
        model.config.rvp.vegetation_classes = [
            VegetationClassesCommand.Record("Veg_Land_HRU", 25, 5.0, 5.0),
            VegetationClassesCommand.Record("Veg_Lake_HRU", 0, 0, 0),
        ]
        model.config.rvp.land_use_classes = [
            LandUseClassesCommand.Record("Landuse_Land_HRU", 0, 1),
            LandUseClassesCommand.Record("Landuse_Lake_HRU", 0, 0),
        ]
        model.config.rvp.channel_profiles = channel_profiles

        #######
        # RVT #
        #######

        streaminputs_extractor = RoutingProductGridWeightExtractor(
            vic_streaminputs_nc_path, routing_product_shp_path
        )

        temperatures_extractor = RoutingProductGridWeightExtractor(
            vic_temperatures_nc_path, routing_product_shp_path
        )

        model.config.rvt.add_nc_variable(
            name="StreamInputs",
            data_type="PRECIP",
            file_name_nc=vic_streaminputs_nc_path.name,
            var_name_nc="Streaminputs",
            dim_names_nc=("lon_dim", "lat_dim", "time"),
            scale=4.0,
            offset=0,
            grid_weights=streaminputs_extractor.extract(),
        )

        model.config.rvt.add_nc_variable(
            name="AverageTemp",
            data_type="TEMP_AVE",
            file_name_nc=vic_temperatures_nc_path.name,
            var_name_nc="Avg_temp",
            dim_names_nc=("lon_dim", "lat_dim", "time"),
            grid_weights=temperatures_extractor.extract(),
        )

        model.config.rvt.add_nc_variable(
            is_observation=True,
            data_type="HYDROGRAPH",
            subbasin_id=gauge.subbasin_id,
            units="m3/s",
            file_name_nc=observation_data_nc_path.name,
            var_name_nc="Q",
            dim_names_nc=("nstations", "time"),
            index=1,  # StationIdx
        )

        #############
        # Run model #
        #############

        model(
            [
                vic_streaminputs_nc_path,
                vic_temperatures_nc_path,
                observation_data_nc_path,
            ],
        )

        ##########
        # Verify #
        ##########

        assert len(model.hydrograph.time) == (end - start).days + 1

        assert model.hydrograph.basin_name.item() == gauge.name

        csv_lines = model.outputs["diagnostics"].read_text().split("\n")
        assert csv_lines[1].split(",")[:-1] == [
            "HYDROGRAPH_ALL",
            observation_data_nc_path.name,
            "0.311049",
            "-11.9514",
            "0.471256",
        ]

        for d, q_sim in [
            (0, 85.97869699520872),
            (1000, 78.9516829670743),
            (2000, 70.29412760595676),
            (3000, 44.711237489482755),
            (4000, 129.98874279175033),
        ]:
            assert model.hydrograph.q_sim[d].item() == pytest.approx(q_sim)



import tempfile

import numpy as np
import pytest

from ravenpy.utilities.testdata import get_local_testdata


class TestHydroBASINS:
    geoserver = pytest.importorskip("ravenpy.utilities.geoserver")

    fiona = pytest.importorskip("fiona")
    gpd = pytest.importorskip("geopandas")
    sgeo = pytest.importorskip("shapely.geometry")

    def test_select_hybas_na_domain(self):
        bbox = (-68.0, 50.0) * 2
        dom = self.geoserver.select_hybas_domain(bbox)
        assert dom == "na"

    def test_select_hybas_ar_domain(self):
        bbox = (-114.65, 61.35) * 2
        dom = self.geoserver.select_hybas_domain(bbox)
        assert dom == "ar"

    def test_get_hydrobasins_location_wfs(self, tmp_path):
        lake_winnipeg = (-98.03575958286369, 52.88238524279493)
        resp = self.geoserver.get_hydrobasins_location_wfs(
            coordinates=lake_winnipeg * 2, domain="na"
        )
        feat = self.gpd.read_file(resp.decode())
        geom = self.sgeo.shape(feat["geometry"][0])
        assert geom.bounds == (-99.2731, 50.3603, -96.2578, 53.8705)
        np.testing.assert_almost_equal(geom.area, 3.2530867)

    def test_get_hydrobasins_attributes_wfs(self, tmp_path):
        rio_grande = (-80.475, 8.4)
        resp = self.geoserver.get_hydrobasins_location_wfs(
            coordinates=rio_grande * 2, domain="na"
        )
        feat = self.gpd.read_file(resp.decode())
        main_bas = feat["MAIN_BAS"][0]

        region_url = self.geoserver.filter_hydrobasins_attributes_wfs(
            attribute="MAIN_BAS", value=main_bas, domain="na"
        )
        gdf = self.gpd.read_file(region_url)

        assert len(gdf) == 18
        assert gdf.crs.to_epsg() == 4326
        assert set(gdf["MAIN_BAS"].values) == {7120000210}

        basin = gdf.dissolve(by="MAIN_BAS")
        np.testing.assert_equal(
            basin.geometry.bounds.values,
            np.array([[-80.8542, 8.2459, -80.1375, 8.7004]]),
        )

    def test_hydrobasins_upstream_aggregate(self, tmp_path):
        puerto_cortes = (-83.525, 8.96)
        resp = self.geoserver.get_hydrobasins_location_wfs(
            coordinates=puerto_cortes * 2, domain="na"
        )
        feat = self.gpd.read_file(resp.decode())

        gdf_upstream = self.geoserver.hydrobasins_upstream(feat.loc[0], domain="na")
        assert len(gdf_upstream) == 73
        aggregated = self.geoserver.hydrobasins_aggregate(gdf_upstream)

        assert len(aggregated) == 1
        assert float(aggregated.SUB_AREA.values) == 4977.8
        np.testing.assert_equal(
            aggregated.geometry.bounds.values,
            np.array([[-83.8167, 8.7625, -82.7125, 9.5875]]),
        )


class TestHydroRouting:
    geoserver = pytest.importorskip("ravenpy.utilities.geoserver")

    fiona = pytest.importorskip("fiona")
    gpd = pytest.importorskip("geopandas")
    sgeo = pytest.importorskip("shapely.geometry")

    def test_hydro_routing_locations(self, tmp_path):
        lake_winnipeg = (-98.03575958286369, 52.88238524279493)
        resp = self.geoserver.get_hydro_routing_location_wfs(
            coordinates=lake_winnipeg * 2, lakes="all"
        )
        feat = self.gpd.read_file(resp.decode())
        geom = feat["geometry"][0]
        assert geom.bounds == (-99.3083, 50.1875, -95.9875, 54.0542)
        # Note: This value is not in sq. km.
        np.testing.assert_almost_equal(geom.area, 4.0978987)

    @pytest.mark.slow
    def test_get_hydro_routing_attributes_wfs(self):
        region_url = self.geoserver.filter_hydro_routing_attributes_wfs(
            attribute="IsLake", value="1.0", lakes="1km", level="07"
        )
        gdf = self.gpd.read_file(region_url)
        assert len(gdf) == 11415

    @pytest.mark.slow
    def test_hydro_routing_upstream(self, tmp_path):
        amadjuak = (-71.225, 65.05)
        resp = self.geoserver.get_hydro_routing_location_wfs(
            coordinates=amadjuak * 2, lakes="1km", level=7
        )
        feature = self.gpd.read_file(resp.decode())
        subbasin_id = feature["SubId"][0]

        gdf_upstream = self.geoserver.hydro_routing_upstream(
            subbasin_id, lakes="1km", level=7
        )

        assert len(gdf_upstream) == 33  # TODO: Verify this with the model maintainers.


class TestWFS:
    geoserver = pytest.importorskip("ravenpy.utilities.geoserver")

    fiona = pytest.importorskip("fiona")
    gpd = pytest.importorskip("geopandas")
    sgeo = pytest.importorskip("shapely.geometry")

    def test_get_location_wfs(self, tmp_path):
        las_vegas = (-115.136389, 36.175)
        usa_admin_bounds = "public:usa_admin_boundaries"
        resp = self.geoserver._get_location_wfs(
            coordinates=las_vegas * 2, layer=usa_admin_bounds
        )
        feat = self.gpd.read_file(resp.decode())

        geom = feat["geometry"][0]
        assert geom.bounds == (-120.001, 35.0019, -114.0417, 41.9948)
        # Note: This value is not in sq. km.
        np.testing.assert_almost_equal(geom.area, 29.9690150)

    def test_get_feature_attributes_wfs(self):
        state_name = "Nevada"
        usa_admin_bounds = "public:usa_admin_boundaries"

        vector_url = self.geoserver._filter_feature_attributes_wfs(
            attribute="STATE_NAME", value=state_name, layer=usa_admin_bounds
        )
        gdf = self.gpd.read_file(vector_url)
        assert len(gdf) == 1
        assert gdf.STATE_NAME.unique() == "Nevada"


class TestWCS:
    io = pytest.importorskip("ravenpy.utilities.io")
    geoserver = pytest.importorskip("ravenpy.utilities.geoserver")
    geo = pytest.importorskip("ravenpy.utilities.geo")
    rasterio = pytest.importorskip("rasterio")

    vector_file = get_local_testdata("polygons/Saskatoon.geojson")

    def test_get_raster_wcs(self, tmp_path):
        # TODO: This CRS needs to be redefined using modern pyproj-compatible strings.
        nalcms_crs = "+proj=laea +lat_0=45 +lon_0=-100 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs=True"

        with tempfile.NamedTemporaryFile(
            prefix="reprojected_", suffix=".json", dir=tmp_path
        ) as projected:
            self.geo.generic_vector_reproject(
                self.vector_file, projected.name, target_crs=nalcms_crs
            )
            bbox = self.io.get_bbox(projected.name)

        raster_url = "public:CEC_NALCMS_LandUse_2010"
        raster_bytes = self.geoserver.get_raster_wcs(
            bbox, geographic=False, layer=raster_url
        )

        with tempfile.NamedTemporaryFile(
            prefix="wcs_", suffix=".tiff", dir=tmp_path
        ) as raster_file:
            with open(raster_file.name, "wb") as rf:
                rf.write(raster_bytes)
                rf.close()
                with self.rasterio.open(rf.name) as src:
                    assert src.width == 650
                    assert src.height == 745
                    np.testing.assert_array_equal(
                        src.lnglat(), (-106.64193764047552, 52.1564202369763)
                    )
                    data = src.read()
                    assert np.unique(data).tolist() == [1, 5, 8, 10, 14, 15, 16, 17, 18]



"""Console script for ravenpy."""

from pathlib import Path

import click

from ravenpy.extractors.routing_product import RoutingProductGridWeightExtractor


@click.command()
@click.argument("input-file", type=click.Path(exists=True))
@click.argument("routing-file", type=click.Path(exists=True))
@click.option(
    "-d",
    "--dim-names",
    nargs=2,
    type=click.Tuple([str, str]),
    default=RoutingProductGridWeightExtractor.DIM_NAMES,
    show_default=True,
    help="Ordered dimension names of longitude (x) and latitude (y) in the NetCDF INPUT_FILE.",
)
@click.option(
    "-v",
    "--var-names",
    nargs=2,
    type=click.Tuple([str, str]),
    default=RoutingProductGridWeightExtractor.VAR_NAMES,
    show_default=True,
    help="Variable name of 1D or 2D longitude and latitude variables in the NetCDF INPUT_FILE (in this order).",
)
@click.option(
    "-c",  # legacy script short option
    "--routing-id-field",
    type=str,
    default=RoutingProductGridWeightExtractor.ROUTING_ID_FIELD,
    show_default=True,
    help="Name of column in routing information shapefile (ROUTING_FILE) containing a unique key for each dataset.",
)
@click.option(
    "-f",  # legacy script short option
    "--netcdf-input-field",
    type=str,
    default=RoutingProductGridWeightExtractor.NETCDF_INPUT_FIELD,
    show_default=True,
    help="Attribute name in INPUT_FILE shapefile that defines the index of the shape in NetCDF model output file (numbering needs to be [0 ... N-1]).",
)
@click.option(
    "-g",
    "--gauge-id",
    "gauge_ids",  # rename arg to put emphasis on the fact that it's a list
    multiple=True,
    type=str,
    show_default=True,
    help="Streamflow gauge IDs of interest (corresponds to 'Gauge_ID' in the ROUTING_FILE shapefile).",
)
@click.option(
    "-s",
    "--sub-id",
    "sub_ids",  # rename arg to put emphasis on the fact that it's a list
    multiple=True,
    type=int,
    show_default=True,
    help="IDs of subbasins of interest (containing usually a gauge station, corresponds to 'SubId' in the ROUTING_FILE shapefile).",
)
@click.option(
    "-e",  # legacy script short option
    "--area-error-threshold",
    type=float,
    default=RoutingProductGridWeightExtractor.AREA_ERROR_THRESHOLD,
    show_default=True,
    help="Threshold (as fraction) of allowed mismatch in areas between subbasins from routing information (ROUTING_FILE) and overlay with grid-cells or subbasins (INPUT_FILE). If error is smaller than this threshold the weights will be adjusted such that they sum up to exactly 1. Raven will exit gracefully in case weights do not sum up to at least 0.95.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(),
    help="Text field that will contain the results as a single :GridWeights Raven command containing the weights.",
)
def generate_grid_weights(
    input_file,
    routing_file,
    dim_names,
    var_names,
    routing_id_field,
    netcdf_input_field,
    gauge_ids,
    sub_ids,
    area_error_threshold,
    output,
):
    """
    Generate grid weights in various formats.

    INPUT_FILE: File containing model discretization. Can be either:

    (A) NetCDF file containing at least 1D or 2D latitudes and 1D or 2D longitudes where this
    grid needs to be representative of model outputs that are then required to be routed. The names of the dimensions
    and the variables holding the lat/lon information should be specified with options --dim-names (-d) and --var-names (-v).

    (B) Shapefile (either a .shp or a .zip) that contains shapes of subbasins and one attribute in this
    shapefile that is defining its index in the NetCDF model output file (numbering needs to be [0 ... N-1]). The name of
    this attribute should be specified via option --netcdf-input-field (-f).

    ROUTING_FILE: Shapefile (either a .shp or a .zip) that contains all information of the routing toolbox for the catchment
    of interest (and maybe some more catchments). This file should contain an attribute with a unique identifier for the HRUs:
    the default is "HRU_ID", but it can be set with --routing-id-field (-c).

    The script will output the results as RVT file with a single :GridWeights command block containing the weights (that the user
    is then free to embed or reference, in her own config context).
    """
    # NOTE: This is in order to make sphinx-click happy. Magic. Do not touch.
    from ravenpy.extractors import RoutingProductGridWeightExtractor

    extractor = RoutingProductGridWeightExtractor(
        input_file,
        routing_file,
        dim_names,
        var_names,
        routing_id_field,
        netcdf_input_field,
        gauge_ids,
        sub_ids,
        area_error_threshold,
    )
    gw_cmd = extractor.extract()

    if not output:
        input_file_path = Path(input_file)
        output_file_path = (
            input_file_path.parent / f"{input_file_path.stem}_weights.rvt"
        )
    else:
        output_file_path = Path(output)

    output_file_path.write_text(gw_cmd.to_rv() + "\n")

    click.echo(f"Created {output_file_path}")



import numpy as np
import xarray as xr
from xclim.indicators.land import fit, stats

from ravenpy.utilities import graphs
from ravenpy.utilities.testdata import get_local_testdata


def test_ts_fit_graph():
    fn = get_local_testdata(
        "hydro_simulations/raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc"
    )
    ds = xr.open_dataset(fn)

    ts = stats(ds.q_sim, op="max", freq="M")
    p = fit(ts)
    np.testing.assert_array_equal(p.isnull(), False)

    fig = graphs.ts_fit_graph(ts, p)
    return fig



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 17 09:11:58 2020

@author: ets
"""

import datetime as dt
import logging
import re
import warnings
from pathlib import Path
from typing import List, Tuple

# import climpred
import numpy as np
import pandas as pd
import xarray as xr
from climpred import HindcastEnsemble

from . import gis_import_error_message

try:
    import rioxarray
    from clisops.core import subset
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

from ravenpy.models.emulators import get_model

LOGGER = logging.getLogger("PYWPS")


# TODO: Complete docstrings

# This function gets model states after running the model (i.e. states at the end of the run).
def get_raven_states(model, workdir=None, **kwds):
    """Get the RAVEN states file (.rvc file) after a model run.

    Parameters
    ----------
    model : {'HMETS', 'GR4JCN', 'MOHYSE', 'HBVEC'}
      Model name.
    kwds : {}
      Model configuration parameters, including the forcing files (ts).

    Returns
    -------
    rvc : {}
      Raven model forcing file

    """
    # Run the model and get the rvc file for future hotstart.
    m = get_model(model)(workdir=workdir)
    m(overwrite=True, **kwds)
    rvc = m.outputs["solution"]

    return rvc


# Do the actual forecasting step
def perform_forecasting_step(rvc, model, workdir=None, **kwds):
    """
    Function that might be useful eventually to do a forecast from a model setup.
    """
    # kwds includes 'ts', the forecast timeseries data
    # Setup the model
    m = get_model(model)(workdir=workdir)

    # Force the initial conditions
    m.resume(rvc)

    # Set the parameters, start dates, etc. required to run the model and run
    m(overwrite=True, **kwds)

    return m.q_sim


def perform_climatology_esp(
    model_name, forecast_date, forecast_duration, workdir=None, **kwds
):
    """
    This function takes the model setup and name as well as forecast data and duration and returns
    an ESP forecast netcdf. The data comes from the climatology data and thus there is a mechanism
    to get the correct data from the time series and exclude the current year.

    Parameters
    ----------
    model_name : {'HMETS', 'MOHYSE', 'GR4JCN', 'HBVEC'}
      Model name to instantiate Raven model.
    forecast_date : datetime.datetime
      Date of the forecast issue.
    forecast_duration : int
      Number of days of forecast, forward looking.
    kwds : dict
      Raven model configuration parameters.

    Returns
    -------
    qsims
      Array of streamflow values from the ESP method along with list of member years

    """
    # Get the timeseries
    tsnc = xr.open_dataset(kwds["ts"])

    # Prepare model instance
    m = get_model(model_name)(workdir=workdir)

    # Now find the periods of time for warm-up and forecast and add to the model keywords as the defaults are failing
    # (nanoseconds datetimes do not like the year 0001...)
    start_date = pd.to_datetime(tsnc["time"][0].values)
    start_date = start_date.to_pydatetime()

    kwds["start_date"] = start_date

    # Forecasting from Feb 29th is not ideal, we will replace with Feb 28th.
    # Should not change much in a climatological forecast.
    if forecast_date.month == 2 and forecast_date.day == 29:
        forecast_date.replace(day=28)

    # Check to make sure forecast date is not in the first year as we need model warm-up.
    # We cannot use timedelta because if the dataset happens to start on a leap
    # year, then the timedelta=365 days will not be robust. (and we cannot use timedelta(years=1)...)
    dateLimit = start_date.replace(year=start_date.year + 1)
    if dateLimit > forecast_date:
        msg = (
            "Forecast date is within the warm-up period. Select another forecast date."
        )
        warnings.warn(msg)

    # initialize the array of forecast variables
    qsims = []

    # list of unique years in the dataset:
    avail_years = list(np.unique(tsnc["time.year"].data))

    # Take a copy of the forecast initial date before overwriting in the forecast step.
    forecast_date_main = forecast_date

    # Remove the year that we are forecasting. Or else it's cheating!
    avail_years.remove(forecast_date.year)

    # Update the forecast end-date, which will be the day prior to the forecast date.
    # So forecasts warm-up will be from day 1 in the dataset to the forecast date.
    kwds["end_date"] = forecast_date - dt.timedelta(days=1)

    # Get RVC file if it exists, else compute it.
    if "rvc" in kwds and len(kwds["rvc"]) > 0:
        rvc = kwds.pop("rvc")
    else:
        # Run model to get rvc file after warm-up using base meteo
        rvc = get_raven_states(model_name, workdir=workdir, **kwds)

    # We need to check which years are long enough (ex: wrapping years, 365-day forecast starting in
    # September 2015 will need data up to August 2016 at least)
    for years in avail_years:
        if forecast_date.replace(year=years) + dt.timedelta(
            days=forecast_duration - 1
        ) > pd.to_datetime(tsnc["time"][-1].values):
            avail_years.remove(years)
            msg = (
                f"Year {years} has been removed because it is the last year in the dataset and does not cover the "
                f"forecast duration."
            )
            warnings.warn(msg)

    # We will iterate this for all forecast years
    for years in avail_years:

        # Replace the forecast period start and end dates with the climatological ESP dates for the
        # current member (year)
        forecast_date = forecast_date.replace(year=years)
        kwds["start_date"] = forecast_date
        kwds["end_date"] = forecast_date + dt.timedelta(days=forecast_duration - 1)

        # Setup the initial states from the warm-up and run the model.
        # Note that info on start/end dates and timeseries are in the kwds.
        m.resume(rvc)
        m(run_name=f"run_{years}", **kwds)

        # Add member to the ensemble and retag the dates to the real forecast dates
        # (or else we will get dates from the climate dataset that cover all years)
        new_member = m.q_sim.copy(deep=True)
        new_member["time"] = pd.date_range(
            forecast_date_main, periods=forecast_duration
        )
        qsims.append(new_member)

    # Concatenate the members through a new dimension for the members and remove unused dims.
    qsims = xr.concat(qsims, dim="member")
    qsims = qsims.squeeze()

    # Add the number of the forecast year as member ID
    qsims["member"] = (["member"], avail_years)

    return qsims


def get_hindcast_day(region_coll, date, climate_model="GEPS"):
    """
    This function generates a forecast dataset that can be used to run raven.
    Data comes from the CASPAR archive and must be aggregated such that each file
    contains forecast data for a single day, but for all forecast timesteps and
    all members.

    The code takes the region shapefile, the forecast date required, and the
    climate_model to use, here GEPS by default, but eventually could be GEPS, GDPS, REPS or RDPS.
    """

    # Get the file locations and filenames as a function of the climate model and date
    [ds, times] = get_CASPAR_dataset(climate_model, date)

    return get_subsetted_forecast(region_coll, ds, times, True)


def get_CASPAR_dataset(climate_model, date):
    """Return Caspar Dataset."""

    if climate_model == "GEPS":
        d = dt.datetime.strftime(date, "%Y%m%d")
        file_url = f"https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/birdhouse/caspar/daily/GEPS_{d}.nc"
        ds = xr.open_dataset(file_url)
        # Here we also extract the times at 6-hour intervals as Raven must have
        # constant timesteps and GEPS goes to 6 hours
        start = pd.to_datetime(ds.time[0].values)
        times = [start + dt.timedelta(hours=n) for n in range(0, 384, 6)]
    else:
        # Eventually: GDPS, RDPS and REPS
        raise NotImplementedError("Only the GEPS model is currently supported")

    # Checking that these exist.
    for f in ["pr", "tas"]:
        if f not in ds:
            raise AttributeError(f"'{f}' not present in dataset")

    return ds, times


def get_ECCC_dataset(climate_model):
    """Return latest GEPS forecast Dataset."""
    if climate_model == "GEPS":
        # Eventually the file will find a permanent home, until then let's use the test folder.
        file_url = "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/dodsC/datasets/forecasts/eccc_geps/GEPS_latest.ncml"

        ds = xr.open_dataset(file_url)
        # Here we also extract the times at 6-hour intervals as Raven must have
        # constant timesteps and GEPS goes to 6 hours
        start = pd.to_datetime(ds.time[0].values)
        times = [start + dt.timedelta(hours=n) for n in range(0, 384, 6)]
    else:
        # Eventually: GDPS, RDPS and REPS
        raise NotImplementedError("Only the GEPS model is currently supported")

    # Checking that these exist. IF the files are still processing, possible that one or both are not available!
    for f in ["pr", "tas"]:
        if f not in ds:
            raise AttributeError(f"'{f}' not present in dataset")

    return ds, times


def get_recent_ECCC_forecast(region_coll, climate_model="GEPS"):
    """
    This function generates a forecast dataset that can be used to run raven.
    Data comes from the ECCC datamart and collected daily. It is aggregated
    such that each file contains forecast data for a single day, but for all
    forecast timesteps and all members.

    The code takes the region shapefile and the climate_model to use, here GEPS
    by default, but eventually could be GEPS, GDPS, REPS or RDPS.

    Parameters
    ----------
    region_coll : fiona.collection.Collection
      The region vectors.
    climate_model : str
      Type of climate model, for now only "GEPS" is supported.

    Returns
    -------
    forecast : xararray.Dataset
      The forecast dataset.

    """

    [ds, times] = get_ECCC_dataset(climate_model)

    return get_subsetted_forecast(region_coll, ds, times, False)


def get_subsetted_forecast(region_coll, ds, times, is_caspar):
    """
    This function takes a dataset, a region and the time sampling array and returns
    the subsetted values for the given region and times

    Parameters
    ----------
    region_coll : fiona.collection.Collection
      The region vectors.
    ds : xarray.Dataset
      The dataset containing the raw, worldwide forecast data
    times: dt.datetime
      The array of times required to do the forecast.
    is_caspar: boolean
      True if the data comes from Caspar, false otherwise. Used to define
      lat/lon on rotated grid.

    Returns
    -------
    forecast : xararray.Dataset
      The forecast dataset.

    """
    # Extract the bounding box to subset the entire forecast grid to something
    # more manageable
    lon_min = region_coll.bounds[0]
    lon_max = region_coll.bounds[2]
    lat_min = region_coll.bounds[1]
    lat_max = region_coll.bounds[3]

    # Subset the data to the desired location (bounding box) and times
    ds = subset.subset_bbox(
        ds, lon_bnds=[lon_min, lon_max], lat_bnds=[lat_min, lat_max]
    ).sel(time=times)

    # Rioxarray requires CRS definitions for variables
    # Get CRS, e.g. 4326
    crs = int(re.match(r"epsg:(\d+)", region_coll.crs["init"]).group(1))

    # Here the name of the variable could differ based on the Caspar file processing
    tas = ds.tas.rio.write_crs(crs)
    pr = ds.pr.rio.write_crs(crs)
    ds = xr.merge([tas, pr])

    # Now apply the mask of the basin contour and average the values to get a single time series
    if is_caspar:
        ds.rio.set_spatial_dims("rlon", "rlat")
        ds["rlon"] = ds["rlon"] - 360
        # clip the netcdf and average across space.
        shdf = [next(iter(region_coll))["geometry"]]
        forecast = ds.rio.clip(shdf, crs=crs)
        forecast = forecast.mean(dim={"rlat", "rlon"}, keep_attrs=True)

    else:
        ds.rio.set_spatial_dims("lon", "lat")
        ds["lon"] = ds["lon"] - 360
        # clip the netcdf and average across space.
        shdf = [next(iter(region_coll))["geometry"]]
        forecast = ds.rio.clip(shdf, crs=crs)
        forecast = forecast.mean(dim={"lat", "lon"}, keep_attrs=True)

    return forecast


def make_climpred_hindcast_object(hindcast, observations):
    """
    This function takes a hindcasting dataset of streamflow as well as associated
    observations and creates a hindcasting object that can be used by the
    climpred toolbox for hindcast verification.

    Parameters
    ----------
    hindcast : xarray.Dataset
      The hindcasting streamflow data for a given period
    observations : xarray.Dataset
      The streamflow observations that are used to verify the hindcasts

    Returns
    -------
    hindcast_obj : climpred.HindcastEnsemble object
      The hindcast ensemble formatted to be used in climpred.

    """

    # Todo: Add verification that the variable names are the same
    # Todo: Add verification of sizes, catch and return message.

    # Make the hindcastEnsemble object for the hindcast data
    hindcast_obj = HindcastEnsemble(hindcast)
    # Add the observations to that hindcastEnsemble object for verification.
    hindcast_obj = hindcast_obj.add_observations(observations)

    return hindcast_obj


def make_ESP_hindcast_dataset(
    model_name, forecast_date, included_years, forecast_duration, **kwargs
) -> Tuple[xr.Dataset, xr.Dataset]:
    """Make a hindcast using ESP dataset.

    This function takes the required information to run a RAVEN model run in ESP
    mode, and runs RAVEN for a series of queried initialization dates. For each
    requested date, it will perform an ESP forecast using all climate data at
    its disposal. Returns the hindcast and observations datasets required by climpred.

    Parameters
    ----------
    model_name: string
      Name of the RAVEN model setup (GR4JCN, MOHYSE, HMETS or HBVEC).
    forecast_date: datetime.datetime
      Calendar date that is used as the base to perform hindcasts.
      The year component is updated for each initialization date.
    included_years: List[int]
      the years that we want to perform a hindcasting for, on the calendar date in "foreacast_date"
    forecast_duration: int
      Duration of the forecast, in days. Refers to the longest lead-time required.
    kwargs: dict
      All other parameters needed to run RAVEN.

    Returns
    -------
    xarray.Dataset
      The dataset containing the (init, member, lead) dimensions ready for using in climpred. (qsim)
    xarray.Dataset
      The dataset containing all observations over the verification period.
      The only dimension is 'time', as required by climpred. No other processing
      is required as climpred will find corresponding verification dates on its own. (qobs)

    Notes
    -----
    init = hindcast issue date
    member = ESP members of the hindcasting experiment
    lead = number of lead days of the forecast.
    """

    # Use the ESP functions to generate the ESP hindcasts for the first period (first initialization)
    qsims = perform_climatology_esp(
        model_name,
        forecast_date.replace(year=included_years[0]),
        forecast_duration,
        **kwargs,
    )

    # climpred needs the 'time' variable name to be 'lead' for the hindcasts.
    # Also add the coordinates for the lead and member dimensions.
    qsims = qsims.rename({"time": "lead"})
    qsims = qsims.assign_coords(lead=list(range(1, forecast_duration + 1)))
    qsims = qsims.assign_coords(member=list(range(1, qsims.data.shape[0] + 1)))

    # Repeat the process for all hindcast years required. Could be parallelized by a pro!
    for i in included_years[1:]:

        qsims_tmp = perform_climatology_esp(
            model_name, forecast_date.replace(year=i), forecast_duration, **kwargs
        )
        qsims_tmp = qsims_tmp.rename({"time": "lead"})
        qsims_tmp = qsims_tmp.assign_coords(lead=list(range(1, forecast_duration + 1)))
        qsims_tmp = qsims_tmp.assign_coords(
            member=list(range(1, qsims_tmp.data.shape[0] + 1))
        )

        # Concatenate in the 'init' dimension.
        qsims = xr.concat([qsims, qsims_tmp], dim="init")

    qsims["lead"] = list(range(1, forecast_duration + 1))

    # Make the list of hindcast dates for populating the 'init' dimension coordinates.
    date_list = [
        forecast_date.replace(year=included_years[x])
        for x in range(len(included_years))
    ]

    # Other processing required by climpred.
    qsims = qsims.assign_coords(init=pd.to_datetime(date_list))
    qsims = qsims.to_dataset(name="flow")

    # Here units are days and always will be!
    qsims["lead"].attrs["units"] = "days"

    # Extract and prepare the observations for verification.
    qobs = xr.open_dataset(kwargs["ts"]).qobs

    # The hindcast and observation variable names MUST be the same. Here we
    # use "flow" but it could be abything else, really.
    qobs = qobs.to_dataset(name="flow")

    return qsims, qobs



import warnings
from collections import defaultdict
from pathlib import Path

try:
    import geopandas
    from osgeo import __version__ as osgeo_version  # noqa
    from osgeo import ogr, osr  # noqa

except (ImportError, ModuleNotFoundError) as e:
    msg = (
        f"`{Path(__file__).stem}` requires installation of the RavenPy GIS libraries. These can be installed using the"
        " `pip install ravenpy[gis]` recipe or via Anaconda (`conda env -n ravenpy-env -f environment.yml`)"
        " from the RavenPy repository source files."
    )
    raise ImportError(msg) from e


import netCDF4 as nc4
import numpy as np

from ravenpy.config.commands import (
    ChannelProfileCommand,
    GridWeightsCommand,
    HRUsCommand,
    ReservoirCommand,
    SubBasinsCommand,
)


class RoutingProductShapefileExtractor:

    """
    This is a class to encapsulate the logic of converting the Routing
    Product into the required data structures to generate the RVH file
    format.
    """

    MAX_RIVER_SLOPE = 0.00001
    USE_LAKE_AS_GAUGE = False
    WEIR_COEFFICIENT = 0.6
    USE_LAND_AS_GAUGE = False
    USE_MANNING_COEFF = False
    MANNING_DEFAULT = 0.035
    HRU_ASPECT_CONVENTION = "GRASS"  # GRASS | ArcGIS

    def __init__(self, shapefile_path, hru_aspect_convention=HRU_ASPECT_CONVENTION):
        if isinstance(shapefile_path, (Path, str)):
            if Path(shapefile_path).suffix == ".zip":
                shapefile_path = f"zip://{shapefile_path}"
            self._df = geopandas.read_file(shapefile_path)
        elif isinstance(shapefile_path, geopandas.GeoDataFrame):
            self._df = shapefile_path

        self.hru_aspect_convention = hru_aspect_convention

    def extract(self, model=None):
        """
        This will extract the data from the Routing Product shapefile and
        return it as relevant Raven command data objects.

        Parameters
        ----------
        model : Raven subclass
          Emulator class used to customize HRU objects extracted.

        Returns
        -------
        dict
            "subbasins"
               Sequence of `commands.SubBasinsCommand.Record` objects
            "land_subbasins"
               Sequence of land subbasins ids
            "lake_subbasins"
               Sequence of lake subbasins ids
            "reservoirs"
               Sequence of `commands.ReservoirCommand` objects
            "channel_profiles"
               Sequence of `commands.ChannelProfileCommand` objects
            "hrus"
               Sequence of `commands.HRUsCommand.Record` objects

        """
        self.model_cls = model

        subbasin_recs = []
        land_sb_ids = []
        lake_sb_ids = []
        reservoir_cmds = []  # those are meant to be injected inline in the RVH
        channel_profile_cmds = []  # those are meant to be injected inline in the RVH
        hru_recs = []

        # Collect all subbasin_ids for fast lookup in next loop
        subbasin_ids = {int(row["SubId"]) for _, row in self._df.iterrows()}

        for _, row in self._df.iterrows():

            # HRU
            hru_recs.append(self._extract_hru(row))

            subbasin_id = int(row["SubId"])

            is_lake = False

            if row["IsLake"] > 0 and row["HRU_IsLake"] > 0:
                lake_sb_ids.append(subbasin_id)
                reservoir_cmds.append(self._extract_reservoir(row))
                is_lake = True
            elif row["IsLake"] > 0:
                continue
            else:
                land_sb_ids.append(subbasin_id)

            # Subbasin
            sb = self._extract_subbasin(row, is_lake, subbasin_ids)
            subbasin_recs.append(sb)

            # ChannelProfile
            channel_profile_cmds.append(self._extract_channel_profile(row))

        return dict(
            subbasins=subbasin_recs,
            land_subbasin_ids=land_sb_ids,
            lake_subbasin_ids=lake_sb_ids,
            reservoirs=reservoir_cmds,
            channel_profiles=channel_profile_cmds,
            hrus=hru_recs,
        )

    def _extract_subbasin(self, row, is_lake, subbasin_ids) -> SubBasinsCommand.Record:
        subbasin_id = int(row["SubId"])
        # is_lake = row["HRU_IsLake"] >= 0
        river_length_in_kms = 0 if is_lake else round(row["Rivlen"] / 1000, 5)
        # river_slope = max(
        #     row["RivSlope"], RoutingProductShapefileExtractor.MAX_RIVER_SLOPE
        # )
        # downstream_id
        downstream_id = int(row["DowSubId"])
        if downstream_id == subbasin_id:
            downstream_id = -1
        elif downstream_id not in subbasin_ids:
            downstream_id = -1
        gauged = row["IsObs"] > 0 or (
            is_lake and RoutingProductShapefileExtractor.USE_LAKE_AS_GAUGE
        )
        rec = SubBasinsCommand.Record(
            subbasin_id=subbasin_id,
            name=f"sub_{subbasin_id}",
            downstream_id=downstream_id,
            profile=f"chn_{subbasin_id}",
            reach_length=river_length_in_kms,
            gauged=gauged,
        )

        return rec

    def _extract_reservoir(self, row) -> ReservoirCommand:
        lake_id = int(row["HyLakeId"])

        return ReservoirCommand(
            subbasin_id=int(row["SubId"]),
            hru_id=int(row["HRU_ID"]),
            name=f"Lake_{lake_id}",
            weir_coefficient=RoutingProductShapefileExtractor.WEIR_COEFFICIENT,
            crest_width=row["BkfWidth"],
            max_depth=row["LakeDepth"],
            lake_area=row["HRU_Area"],
        )

    def _extract_channel_profile(self, row) -> ChannelProfileCommand:
        subbasin_id = int(row["SubId"])
        slope = max(row["RivSlope"], RoutingProductShapefileExtractor.MAX_RIVER_SLOPE)

        # SWAT: top width of channel when filled with water; bankfull width W_bnkfull
        channel_width = max(row["BkfWidth"], 1)
        # SWAT: depth of water in channel when filled to top of bank
        channel_depth = max(row["BkfDepth"], 1)
        channel_elev = row["MeanElev"]
        floodn = row["FloodP_n"]
        channeln = row["Ch_n"]

        # channel profile calculations are based on theory SWAT model is based on
        # see: https://swat.tamu.edu/media/99192/swat2009-theory.pdf
        #      --> "Channel Characteristics" p. 429 ff

        # inverse of channel side slope; channel sides assumed to have 2:1 run to rise ratio
        zch = 2
        # river side width
        sidwd = zch * channel_depth
        # river bottom width W_btm
        botwd = channel_width - 2 * sidwd

        # if derived bottom width is negative, set bottom width to 0.5*bankfull width and recalculate zch
        if botwd < 0:
            botwd = 0.5 * channel_width
            sidwd = 0.5 * 0.5 * channel_width
            zch = (channel_width - botwd) / (2 * channel_depth)

        # inverse of floodplain side slope; flood plain side slopes assumed to have 4:1 run to rise ratio
        zfld = 4 + channel_elev
        # floodplain bottom width
        zbot = channel_elev - channel_depth
        # floodplain side width
        sidwdfp = 4 / 0.25

        # geometry of the channel and floodplain
        # (see figure 7:1-2 in SWAT theory document)
        survey_points = [
            (0, zfld),
            (sidwdfp, channel_elev),
            (sidwdfp + 2 * channel_width, channel_elev),
            (sidwdfp + 2 * channel_width + sidwd, zbot),
            (sidwdfp + 2 * channel_width + sidwd + botwd, zbot),
            (sidwdfp + 2 * channel_width + 2 * sidwd + botwd, channel_elev),
            (sidwdfp + 4 * channel_width + 2 * sidwd + botwd, channel_elev),
            (2 * sidwdfp + 4 * channel_width + 2 * sidwd + botwd, zfld),
        ]

        if RoutingProductShapefileExtractor.USE_MANNING_COEFF:
            mann = channeln
        else:
            mann = RoutingProductShapefileExtractor.MANNING_DEFAULT

        # roughness zones of channel and floodplain
        roughness_zones = [
            (0, floodn),
            (sidwdfp + 2 * channel_width, mann),
            (sidwdfp + 2 * channel_width + 2 * sidwd + botwd, floodn),
        ]

        return ChannelProfileCommand(
            name=f"chn_{subbasin_id}",
            bed_slope=slope,
            survey_points=survey_points,
            roughness_zones=roughness_zones,
        )

    def _extract_hru(self, row) -> HRUsCommand.Record:

        aspect = row["HRU_A_mean"]

        if self.hru_aspect_convention == "GRASS":
            aspect -= 360
            if aspect < 0:
                aspect += 360
        elif self.hru_aspect_convention == "ArcGIS":
            aspect = 360 - aspect
        else:
            assert False

        attrs = dict(
            hru_id=int(row["HRU_ID"]),
            area=row["HRU_Area"] / 1_000_000,
            elevation=row["HRU_E_mean"],
            latitude=row["HRU_CenY"],
            longitude=row["HRU_CenX"],
            subbasin_id=int(row["SubId"]),
            aquifer_profile="[NONE]",
            terrain_class="[NONE]",
            slope=row["HRU_S_mean"],
            aspect=aspect,
        )

        if self.model_cls is not None:
            # Instantiate HRUs with emulator specific land_use_class, veg_class and soil_profile names.
            if row["LAND_USE_C"] == "Landuse_Land_HRU":
                return self.model_cls.LandHRU(**attrs)

            if row["LAND_USE_C"] == "Landuse_Lake_HRU":
                return self.model_cls.LakeHRU(**attrs)

        # Instantiate HRUs with generic land_use_class, veg_class and soil_profile names.
        return HRUsCommand.Record(
            land_use_class=row["LAND_USE_C"],
            veg_class=row["VEG_C"],
            soil_profile=row["SOIL_PROF"],
            **attrs,
        )


class RoutingProductGridWeightExtractor:

    """
    The original version of this algorithm can be found at: https://github.com/julemai/GridWeightsGenerator
    """

    DIM_NAMES = ("lon_dim", "lat_dim")
    VAR_NAMES = ("lon", "lat")
    ROUTING_ID_FIELD = "HRU_ID"
    NETCDF_INPUT_FIELD = "NetCDF_col"
    AREA_ERROR_THRESHOLD = 0.05
    CRS_LLDEG = 4326  # EPSG id of lat/lon (deg) coordinate reference system (CRS)
    CRS_CAEA = 3573  # EPSG id of equal-area coordinate reference system (CRS)

    def __init__(
        self,
        input_file_path,
        routing_file_path,
        dim_names=DIM_NAMES,
        var_names=VAR_NAMES,
        routing_id_field=ROUTING_ID_FIELD,
        netcdf_input_field=NETCDF_INPUT_FIELD,
        gauge_ids=None,
        sub_ids=None,
        area_error_threshold=AREA_ERROR_THRESHOLD,
    ):
        self._dim_names = tuple(dim_names)
        self._var_names = tuple(var_names)
        self._routing_id_field = routing_id_field
        self._netcdf_input_field = netcdf_input_field
        self._gauge_ids = gauge_ids or []
        self._sub_ids = sub_ids or []
        self._area_error_threshold = area_error_threshold

        assert not (
            self._gauge_ids and self._sub_ids
        ), "Only one of gauge_ids or sub_ids can be specified"

        self._input_is_netcdf = True

        input_file_path = Path(input_file_path)
        if input_file_path.suffix == ".nc":
            # Note that we cannot use xarray because it complains about variables and dimensions
            # having the same name.
            self._input_data = nc4.Dataset(input_file_path)
        elif input_file_path.suffix == ".zip":
            self._input_data = geopandas.read_file(f"zip://{input_file_path}")
            self._input_is_netcdf = False
        elif input_file_path.suffix == ".shp":
            self._input_data = geopandas.read_file(input_file_path)
            self._input_is_netcdf = False
        else:
            raise ValueError(
                "The input file must be a shapefile (.shp or .zip) or NetCDF"
            )

        routing_file_path = Path(routing_file_path)
        if routing_file_path.suffix == ".zip":
            routing_file_path = f"zip://{routing_file_path}"
        self._routing_data = geopandas.read_file(routing_file_path)

    def extract(self) -> GridWeightsCommand:
        self._prepare_input_data()

        # Read routing data

        # WGS 84 / North Pole LAEA Canada
        self._routing_data = self._routing_data.to_crs(
            epsg=RoutingProductGridWeightExtractor.CRS_CAEA
        )

        def keep_only_valid_downsubid_and_obs_nm(g):
            """
            This function receives a group (g) of routing rows that have been grouped by HRU_ID.
            If there is only one row, there is nothing to do (we return it). If there are more than
            one rows, we want to return the one with the DowSubId != -1 (if there are more than one
            we emit a warning). We also want to search for an Obs_NM value different than -9999 in
            any row of the group, and if we find it we set it in the corresponding field of the
            returning row (if we can't find one we also emit a warning).
            """
            if len(g) == 1:
                return g
            hru_id_field = self._routing_id_field
            row = g[g["DowSubId"] != -1].copy()
            if len(row) > 1:
                row = row[:1].copy()
                warnings.warn(
                    f"More than one row with HRU_ID={row[hru_id_field]} having DowSubId = -1"
                )
            obs_nm = g[g["Obs_NM"] != -9999]
            if not obs_nm.empty:
                row["Obs_NM"] = obs_nm["Obs_NM"].iloc[0]
            else:
                warnings.warn(
                    f"All values of Obs_NM are -9999 for rows with HRU_ID={row[hru_id_field]}"
                )

            return row

        # Remove duplicate HRU_IDs while making sure that we keed relevant DowSubId and Obs_NM values
        self._routing_data = self._routing_data.groupby(self._routing_id_field).apply(
            keep_only_valid_downsubid_and_obs_nm
        )

        # Make sure those are ints
        self._routing_data.SubId = self._routing_data.SubId.astype(int)
        self._routing_data.DowSubId = self._routing_data.DowSubId.astype(int)

        if self._gauge_ids:
            # Extract the SubIDs of the gauges that were specified at input
            self._sub_ids = (
                self._routing_data.loc[self._routing_data.Obs_NM.isin(self._gauge_ids)]
                .SubId.unique()
                .tolist()
            )
            if not self._sub_ids:
                raise ValueError(
                    f"No shapes were found with gauge ID (Obs_NM) in {self._gauge_ids}"
                )

        if self._sub_ids:
            # Here we want to extract the network of connected subbasins by going upstream via their DowSubId,
            # starting from the list supplied by the user (either directly, or via their gauge IDs).. We first
            # build a map of downSubID -> subID for efficient lookup
            downsubid_to_subids = defaultdict(set)
            for _, r in self._routing_data.iterrows():
                downsubid_to_subids[r.DowSubId].add(r.SubId)

            expanded_sub_ids = set(self._sub_ids)
            prev = expanded_sub_ids.copy()
            while True:
                curr = set()
                for did in prev:
                    curr |= downsubid_to_subids[did]
                if curr:
                    expanded_sub_ids |= curr
                    prev = curr
                else:
                    break

            self._sub_ids = expanded_sub_ids

        # Reduce the initial dataset with the target Sub IDs
        if self._sub_ids:
            self._routing_data = self._routing_data[
                self._routing_data.SubId.isin(self._sub_ids)
            ]

        # -------------------------------
        # construct all grid cell polygons
        # -------------------------------

        grid_cell_geom_gpd_wkt = self._compute_grid_cell_polygons()

        # -------------------------------
        # Derive overlay and calculate weights
        # -------------------------------

        grid_weights = []

        for _, row in self._routing_data.iterrows():

            poly = ogr.CreateGeometryFromWkt(row.geometry.to_wkt())

            area_basin = poly.Area()
            # bounding box around basin (for easy check of proximity)
            enve_basin = poly.GetEnvelope()

            area_all = 0.0
            # ncells = 0

            row_grid_weights = []

            for ilat in range(self._nlat):
                for ilon in range(self._nlon):

                    # bounding box around grid-cell (for easy check of proximity)
                    enve_gridcell = grid_cell_geom_gpd_wkt[ilat][ilon].GetEnvelope()

                    grid_is_close = self._check_proximity_of_envelops(
                        enve_gridcell, enve_basin
                    )

                    # this check decreases runtime DRASTICALLY (from ~6h to ~1min)
                    if not grid_is_close:
                        continue

                    # grid_cell_area = grid_cell_geom_gpd_wkt[ilat][ilon].Area()

                    # "fake" buffer to avoid invalid polygons and weirdos dumped by ArcGIS
                    inter = grid_cell_geom_gpd_wkt[ilat][ilon].Intersection(
                        poly.Buffer(0.0)
                    )

                    area_intersect = inter.Area()

                    area_all += area_intersect

                    if area_intersect > 0:
                        hru_id = int(row[self._routing_id_field])
                        cell_id = ilat * self._nlon + ilon
                        weight = area_intersect / area_basin
                        row_grid_weights.append((hru_id, cell_id, weight))

            # mismatch between area of subbasin (routing product) and sum of all contributions of grid cells (model output)
            error = (area_basin - area_all) / area_basin

            if abs(error) > self._area_error_threshold and area_basin > 500000.0:
                # record all basins with errors larger 5% (if basin is larger than 0.5 km2)
                # error_dict[int(ibasin[key_colname])] = [error, area_basin]
                grid_weights += row_grid_weights

            else:
                # adjust such that weights sum up to 1.0
                for hru_id, cell_id, weight in row_grid_weights:
                    corrected_weight = weight * 1.0 / (1.0 - error)
                    grid_weights.append((hru_id, cell_id, corrected_weight))

                # if error < 1.0:
                #     area_all *= 1.0 / (1.0 - error)
                # error = 0.0

        return GridWeightsCommand(
            number_hrus=len(self._routing_data),
            number_grid_cells=self._nlon * self._nlat,
            data=grid_weights,
        )

    def _prepare_input_data(self):

        if self._input_is_netcdf:

            # Raven numbering is:
            #
            #      [      1      2      3   ...     1*nlon
            #        nlon+1 nlon+2 nlon+3   ...     2*nlon
            #           ...    ...    ...   ...     ...
            #           ...    ...    ...   ...  nlat*nlon ]
            #
            # --> Making sure shape of lat/lon fields is like that
            #

            lon_var = self._input_data.variables[self._var_names[0]]
            lon_dims = lon_var.dimensions
            lon_var = lon_var[:]

            lat_var = self._input_data.variables[self._var_names[1]]
            lat_dims = lat_var.dimensions
            lat_var = lat_var[:]

            if len(lon_dims) == 2 and len(lat_dims) == 2:
                if lon_dims == self._dim_names:
                    lon_var = np.transpose(lon_var)

                if lat_dims == self._dim_names:
                    lat_var = np.transpose(lat_var)

            elif len(lon_dims) == 1 and len(lat_dims) == 1:
                # Coord vars are 1d, make them 2d
                lon_var_size = lon_var.size
                lon_var = np.tile(lon_var, (lat_var.size, 1))
                lat_var = np.transpose(np.tile(lat_var, (lon_var_size, 1)))
                assert lon_var.shape == lat_var.shape

            else:
                raise ValueError(
                    "The coord variables of the input data must have the same number of dimensions (either 1 or 2)"
                )

            self._lath, self._lonh = self._create_gridcells_from_centers(
                lat_var, lon_var
            )

            self._nlon = np.shape(lon_var)[1]
            self._nlat = np.shape(lat_var)[0]

        else:

            # input data is a shapefile

            self._input_data = self._input_data.to_crs(
                epsg=RoutingProductGridWeightExtractor.CRS_CAEA
            )

            self._nlon = 1  # only for consistency

            # number of shapes in model "discretization" shapefile (not routing toolbox shapefile)
            self._nlat = self._input_data.geometry.count()  # only for consistency

    def _compute_grid_cell_polygons(self):

        grid_cell_geom_gpd_wkt = [
            [[] for ilon in range(self._nlon)] for ilat in range(self._nlat)
        ]

        if self._input_is_netcdf:

            lath = self._lath
            lonh = self._lonh

            for ilat in range(self._nlat):

                for ilon in range(self._nlon):

                    # -------------------------
                    # EPSG:3035   needs a swap before and after transform ...
                    # -------------------------
                    # gridcell_edges = [ [lath[ilat,ilon]    , lonh[ilat,  ilon]    ],            # for some reason need to switch lat/lon that transform works
                    #                    [lath[ilat+1,ilon]  , lonh[ilat+1,ilon]    ],
                    #                    [lath[ilat+1,ilon+1], lonh[ilat+1,ilon+1]  ],
                    #                    [lath[ilat,ilon+1]  , lonh[ilat,  ilon+1]  ]]

                    # tmp = self._routing_data_to_geometry(gridcell_edges, epsg=crs_caea)
                    # tmp.SwapXY()              # switch lat/lon back
                    # grid_cell_geom_gpd_wkt[ilat][ilon] = tmp

                    # -------------------------
                    # EPSG:3573   does not need a swap after transform ... and is much faster than transform with EPSG:3035
                    # -------------------------
                    #
                    # Windows            Python 3.8.5 GDAL 3.1.3 --> lat/lon (Ming)
                    # MacOS 10.15.6      Python 3.8.5 GDAL 3.1.3 --> lat/lon (Julie)
                    # Graham             Python 3.8.2 GDAL 3.0.4 --> lat/lon (Julie)
                    # Graham             Python 3.6.3 GDAL 2.2.1 --> lon/lat (Julie)
                    # Ubuntu 18.04.2 LTS Python 3.6.8 GDAL 2.2.3 --> lon/lat (Etienne)
                    #
                    if osgeo_version < "3.0":
                        gridcell_edges = [
                            [
                                lonh[ilat, ilon],
                                lath[ilat, ilon],
                            ],  # for some reason need to switch lat/lon that transform works
                            [lonh[ilat + 1, ilon], lath[ilat + 1, ilon]],
                            [lonh[ilat + 1, ilon + 1], lath[ilat + 1, ilon + 1]],
                            [lonh[ilat, ilon + 1], lath[ilat, ilon + 1]],
                        ]
                    else:
                        gridcell_edges = [
                            [
                                lath[ilat, ilon],
                                lonh[ilat, ilon],
                            ],  # for some reason lat/lon order works
                            [lath[ilat + 1, ilon], lonh[ilat + 1, ilon]],
                            [lath[ilat + 1, ilon + 1], lonh[ilat + 1, ilon + 1]],
                            [lath[ilat, ilon + 1], lonh[ilat, ilon + 1]],
                        ]

                    tmp = self._shape_to_geometry(
                        gridcell_edges, epsg=RoutingProductGridWeightExtractor.CRS_CAEA
                    )
                    grid_cell_geom_gpd_wkt[ilat][ilon] = tmp

        else:

            for ishape in range(self._nlat):

                idx = np.where(self._input_data[self._netcdf_input_field] == ishape)[0]
                if len(idx) == 0:
                    # print(
                    #     "Polygon ID = {} not found in '{}'. Numbering of shapefile attribute '{}' needs to be [0 ... {}-1].".format(
                    #         ishape, input_file, key_colname_model, nshapes
                    #     )
                    # )
                    raise ValueError("Polygon ID not found.")
                if len(idx) > 1:
                    # print(
                    #     "Polygon ID = {} found multiple times in '{}' but needs to be unique. Numbering of shapefile attribute '{}' needs to be [0 ... {}-1].".format(
                    #         ishape, input_file, key_colname_model, nshapes
                    #     )
                    # )
                    raise ValueError("Polygon ID not unique.")
                idx = idx[0]
                poly = self._input_data.loc[idx].geometry
                grid_cell_geom_gpd_wkt[ishape][0] = ogr.CreateGeometryFromWkt(
                    poly.to_wkt()
                ).Buffer(
                    0.0
                )  # We add an empty buffer here to fix problems with bad polygon topology (actually caused by ESRI's historical incompetence)

        return grid_cell_geom_gpd_wkt

    def _create_gridcells_from_centers(self, lat, lon):

        # create array of edges where (x,y) are always center cells
        nlon = np.shape(lon)[1]
        nlat = np.shape(lat)[0]
        lonh = np.empty((nlat + 1, nlon + 1), dtype=np.float)
        lath = np.empty((nlat + 1, nlon + 1), dtype=np.float)
        tmp1 = [
            [(lat[ii + 1, jj + 1] - lat[ii, jj]) / 2 for jj in range(nlon - 1)]
            + [(lat[ii + 1, nlon - 1] - lat[ii, nlon - 2]) / 2]
            for ii in range(nlat - 1)
        ]
        tmp2 = [
            [(lon[ii + 1, jj + 1] - lon[ii, jj]) / 2 for jj in range(nlon - 1)]
            + [(lon[ii + 1, nlon - 1] - lon[ii, nlon - 2]) / 2]
            for ii in range(nlat - 1)
        ]
        dlat = np.array(tmp1 + [tmp1[-1]])
        dlon = np.array(tmp2 + [tmp2[-1]])
        lonh[0:nlat, 0:nlon] = lon - dlon
        lath[0:nlat, 0:nlon] = lat - dlat

        # make lat and lon one column and row wider such that all
        lonh[nlat, 0:nlon] = lonh[nlat - 1, 0:nlon] + (
            lonh[nlat - 1, 0:nlon] - lonh[nlat - 2, 0:nlon]
        )
        lath[nlat, 0:nlon] = lath[nlat - 1, 0:nlon] + (
            lath[nlat - 1, 0:nlon] - lath[nlat - 2, 0:nlon]
        )
        lonh[0:nlat, nlon] = lonh[0:nlat, nlon - 1] + (
            lonh[0:nlat, nlon - 1] - lonh[0:nlat, nlon - 2]
        )
        lath[0:nlat, nlon] = lath[0:nlat, nlon - 1] + (
            lath[0:nlat, nlon - 1] - lath[0:nlat, nlon - 2]
        )
        lonh[nlat, nlon] = lonh[nlat - 1, nlon - 1] + (
            lonh[nlat - 1, nlon - 1] - lonh[nlat - 2, nlon - 2]
        )
        lath[nlat, nlon] = lath[nlat - 1, nlon - 1] + (
            lath[nlat - 1, nlon - 1] - lath[nlat - 2, nlon - 2]
        )

        return [lath, lonh]

    def _shape_to_geometry(self, shape_from_jsonfile, epsg=None):

        # converts shape read from shapefile to geometry
        # epsg :: integer EPSG code

        ring_shape = ogr.Geometry(ogr.wkbLinearRing)

        for ii in shape_from_jsonfile:
            ring_shape.AddPoint_2D(ii[0], ii[1])
        # close ring
        ring_shape.AddPoint_2D(shape_from_jsonfile[0][0], shape_from_jsonfile[0][1])

        poly_shape = ogr.Geometry(ogr.wkbPolygon)
        poly_shape.AddGeometry(ring_shape)

        if epsg:
            source = osr.SpatialReference()
            # usual lat/lon projection
            source.ImportFromEPSG(RoutingProductGridWeightExtractor.CRS_LLDEG)

            target = osr.SpatialReference()
            target.ImportFromEPSG(epsg)  # any projection to convert to

            transform = osr.CoordinateTransformation(source, target)
            poly_shape.Transform(transform)

        return poly_shape

    def _check_proximity_of_envelops(self, gridcell_envelop, shape_envelop):

        # checks if two envelops are in proximity (intersect)

        # minX  --> env[0]
        # maxX  --> env[1]
        # minY  --> env[2]
        # maxY  --> env[3]

        return (
            gridcell_envelop[0] <= shape_envelop[1]
            and gridcell_envelop[1] >= shape_envelop[0]
            and gridcell_envelop[2] <= shape_envelop[3]
            and gridcell_envelop[3] >= shape_envelop[2]
        )

    def _check_gridcell_in_proximity_of_shape(
        self, gridcell_edges, shape_from_jsonfile
    ):

        # checks if a grid cell falls into the bounding box of the shape
        # does not mean it intersects but it is a quick and cheap way to
        # determine cells that might intersect

        # gridcell_edges = [(lon1,lat1),(lon2,lat2),(lon3,lat3),(lon4,lat4)]
        # shape_from_jsonfile

        min_lat_cell = np.min([ii[1] for ii in gridcell_edges])
        max_lat_cell = np.max([ii[1] for ii in gridcell_edges])
        min_lon_cell = np.min([ii[0] for ii in gridcell_edges])
        max_lon_cell = np.max([ii[0] for ii in gridcell_edges])

        lat_shape = np.array(
            [icoord[1] for icoord in shape_from_jsonfile]
        )  # is it lat???
        lon_shape = np.array(
            [icoord[0] for icoord in shape_from_jsonfile]
        )  # is it lon???

        min_lat_shape = np.min(lat_shape)
        max_lat_shape = np.max(lat_shape)
        min_lon_shape = np.min(lon_shape)
        max_lon_shape = np.max(lon_shape)

        return (
            min_lat_cell <= max_lat_shape
            and max_lat_cell >= min_lat_shape
            and min_lon_cell <= max_lon_shape
            and max_lon_cell >= min_lon_shape
        )



import pytest
import xarray as xr
from xclim.indicators.land._streamflow import fit, stats

from ravenpy.utilities.testdata import get_local_testdata


@pytest.fixture
def q_sim_1(tmp_path):
    """A file storing a Raven streamflow simulation over one basin."""
    return get_local_testdata(
        "hydro_simulations/raven-gr4j-cemaneige-sim_hmets-0_Hydrographs.nc"
    )


@pytest.fixture
def ts_stats(q_sim_1, tmp_path):
    q = xr.open_dataset(q_sim_1).q_sim
    ts = stats(q, op="max")
    fn = tmp_path / "ts_stats.nc"
    ts.to_netcdf_(fn)
    return fn


@pytest.fixture
def params(ts_stats, tmp_path):
    ds = xr.open_dataset(ts_stats)
    name = list(ds.data_vars.keys()).pop()
    q = ds[name]
    p = fit(q, dist="gumbel_r")
    fn = tmp_path / "fit.nc"
    p.to_netcdf(fn)
    return fn



"""
Tools for searching for and acquiring test data.
"""
import logging
import os
import re
from pathlib import Path
from typing import List, Optional, Sequence, Union
from urllib.error import HTTPError
from urllib.parse import urljoin
from urllib.request import urlretrieve

import requests
from xarray import Dataset
from xarray import open_dataset as _open_dataset
from xarray.tutorial import file_md5_checksum

_default_cache_dir = Path.home() / ".raven_testing_data"

LOGGER = logging.getLogger("RAVEN")

__all__ = ["get_local_testdata", "get_file", "open_dataset", "query_folder"]


def get_local_testdata(pattern: str) -> Union[Path, List[Path]]:
    """Gather testdata from a local folder.

    Return files matching `pattern` in the local test data repo
    located at `RAVENPY_TESTDATA_PATH` (which must be set).

    Parameters
    ----------
    pattern: str
      Glob pattern, which must include the folder.

    Returns
    -------
    Union[Path, List[Path]]
    """
    testdata_path = os.getenv("RAVENPY_TESTDATA_PATH")
    if not testdata_path:
        raise RuntimeError("RAVENPY_TESTDATA_PATH env variable is not set")
    testdata_path = Path(testdata_path)
    if not testdata_path.exists():
        raise RuntimeError(f"{testdata_path} does not exists")
    paths = [path for path in testdata_path.glob(pattern) if path.suffix != ".md5"]
    if not paths:
        raise RuntimeError(f"No data found for {pattern} at {testdata_path}")
    # Return item directly when singleton, for convenience
    return paths[0] if len(paths) == 1 else paths


def _get(
    fullname: Path,
    github_url: str,
    branch: str,
    suffix: str,
    cache_dir: Path,
) -> Path:
    cache_dir = cache_dir.absolute()
    local_file = cache_dir / branch / fullname
    md5name = fullname.with_suffix("{}.md5".format(suffix))
    md5file = cache_dir / branch / md5name

    if not local_file.is_file():
        # This will always leave this directory on disk.
        # We may want to add an option to remove it.
        local_file.parent.mkdir(parents=True, exist_ok=True)

        url = "/".join((github_url, "raw", branch, fullname.as_posix()))
        LOGGER.info("Fetching remote file: %s" % fullname.as_posix())
        urlretrieve(url, local_file)
        try:
            url = "/".join((github_url, "raw", branch, md5name.as_posix()))
            LOGGER.info("Fetching remote file md5: %s" % md5name.as_posix())
            urlretrieve(url, md5file)
        except HTTPError as e:
            msg = f"{md5name.as_posix()} not found. Aborting file retrieval."
            local_file.unlink()
            raise FileNotFoundError(msg) from e

        localmd5 = file_md5_checksum(local_file)
        try:
            with open(md5file) as f:
                remotemd5 = f.read()
            if localmd5 != remotemd5:
                local_file.unlink()
                msg = """
                    MD5 checksum does not match, try downloading dataset again.
                    """
                raise OSError(msg)
        except OSError as e:
            LOGGER.error(e)
    return local_file


# idea copied from xclim that borrowed it from xarray that was borrowed from Seaborn
def get_file(
    name: Union[str, Sequence[str]],
    github_url: str = "https://github.com/Ouranosinc/raven-testdata",
    branch: str = "master",
    cache_dir: Path = _default_cache_dir,
) -> Union[Path, List[Path]]:
    """
    Return a file from an online GitHub-like repository.
    If a local copy is found then always use that to avoid network traffic.

    Parameters
    ----------
    name : Union[str, Sequence[str]]
        Name of the file or list/tuple of names of files containing the dataset(s) including suffixes.
    github_url : str
        URL to Github repository where the data is stored.
    branch : str, optional
        For GitHub-hosted files, the branch to download from.
    cache_dir : Path
        The directory in which to search for and write cached data.

    Returns
    -------
    Union[Path, List[Path]]
    """
    if isinstance(name, str):
        name = [name]

    files = list()
    for n in name:
        fullname = Path(n)
        suffix = fullname.suffix
        files.append(
            _get(
                fullname=fullname,
                github_url=github_url,
                branch=branch,
                suffix=suffix,
                cache_dir=cache_dir,
            )
        )
    if len(files) == 1:
        return files[0]
    return files


# Credits to Anselme  https://stackoverflow.com/a/62003257/7322852 (CC-BY-SA 4.0)
def query_folder(
    folder: Optional[str] = None,
    pattern: Optional[str] = None,
    github_url: str = "https://github.com/Ouranosinc/raven-testdata",
    branch: str = "master",
) -> List[str]:
    """
    Lists the files available for retrieval from a remote git repository with get_file.
    If provided a folder name, will perform a globbing-like filtering operation for parent folders.

    Parameters
    ----------
    folder : str, optional
        Relative pathname of the sub-folder from the top-level.
    pattern : str, optional
        Regex pattern to identify a file.
    github_url : str
        URL to Github repository where the data is stored.
    branch : str, optional
        For GitHub-hosted files, the branch to download from.

    Returns
    -------
    List[str]
    """
    repo_name = github_url.strip("https://github.com/")

    url = f"https://api.github.com/repos/{repo_name}/git/trees/{branch}?recursive=1"
    r = requests.get(url)
    res = r.json()

    md5_files = [f["path"] for f in res["tree"] if f["path"].endswith(".md5")]
    if folder:
        folder = "/".join("/".split(folder)) if "/" in folder else folder
        md5_files = [f for f in md5_files if folder in Path(f).parent.as_posix()]
    files = [re.sub(".md5$", "", f) for f in md5_files]

    if pattern:
        regex = re.compile(pattern)
        files = [string for string in files if re.search(regex, string)]

    return files


# idea copied from xclim that borrowed it from xarray that was borrowed from Seaborn
def open_dataset(
    name: str,
    suffix: Optional[str] = None,
    dap_url: Optional[str] = None,
    github_url: str = "https://github.com/Ouranosinc/raven-testdata",
    branch: str = "master",
    cache: bool = True,
    cache_dir: Path = _default_cache_dir,
    **kwds,
) -> Dataset:
    """Open a dataset from the online GitHub-like repository.

    If a local copy is found then always use that to avoid network traffic.

    Parameters
    ----------
    name: str
      Name of the file containing the dataset. If no suffix is given, assumed to be netCDF ('.nc' is appended).
    suffix: str, optional
      If no suffix is given, assumed to be netCDF ('.nc' is appended). For no suffix, set "".
    dap_url: str, optional
      URL to OPeNDAP folder where the data is stored. If supplied, supersedes github_url.
    github_url: str
      URL to Github repository where the data is stored.
    branch: str, optional
      For GitHub-hosted files, the branch to download from.
    cache_dir: Path
      The directory in which to search for and write cached data.
    cache: bool
      If True, then cache data locally for use on subsequent calls.
    kwds: dict, optional
      For NetCDF files, keywords passed to xarray.open_dataset.

    Returns
    -------
    Union[Dataset, Path]

    See Also
    --------
    xarray.open_dataset
    """
    name = Path(name)
    if suffix is None:
        suffix = ".nc"
    fullname = name.with_suffix(suffix)

    if dap_url is not None:
        dap_file = urljoin(dap_url, str(name))
        try:
            ds = _open_dataset(dap_file, **kwds)
            return ds
        except OSError:
            msg = "OPeNDAP file not read. Verify that service is available."
            LOGGER.error(msg)
            raise

    local_file = _get(
        fullname=fullname,
        github_url=github_url,
        branch=branch,
        suffix=suffix,
        cache_dir=cache_dir,
    )

    try:
        ds = _open_dataset(local_file, **kwds)
        if not cache:
            ds = ds.load()
            local_file.unlink()
        return ds
    except OSError:
        raise



"""
Tools for reading and writing geospatial data formats.
"""
import logging
import tarfile
import tempfile
import warnings
import zipfile
from pathlib import Path
from re import search
from typing import Iterable, List, Optional, Sequence, Union

from . import gis_import_error_message

try:
    import fiona
    import rasterio
    from pyproj import CRS
    from shapely.geometry import shape
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

LOGGER = logging.getLogger("RavenPy")
WGS84 = 4326


def address_append(address: Union[str, Path]) -> str:
    """
    Formats a URL/URI to be more easily read with libraries such as "rasterstats"

    Parameters
    ----------
    address: Union[str, Path]
      URL/URI to a potential zip or tar file

    Returns
    -------
    str
      URL/URI prefixed for archive type
    """
    zipped = search(r"(\.zip)", str(address))
    tarred = search(r"(\.tar)", str(address))

    try:
        if zipped:
            return f"zip://{address}"
        elif tarred:
            return f"tar://{address}"
        else:
            LOGGER.info("No prefixes needed for address.")
            return str(address)
    except Exception:
        LOGGER.error("Failed to prefix or parse URL %s." % address)
        raise


def generic_extract_archive(
    resources: Union[str, Path, List[Union[bytes, str, Path]]],
    output_dir: Optional[Union[str, Path]] = None,
) -> List[str]:
    """Extract archives (tar/zip) to a working directory.

    Parameters
    ----------
    resources: Union[str, Path, List[Union[bytes, str, Path]]]
      list of archive files (if netCDF files are in list, they are passed and returned as well in the return).
    output_dir: Optional[Union[str, Path]]
      string or Path to a working location (default: temporary folder).

    Returns
    -------
    list
      List of original or of extracted files
    """

    archive_types = [".tar", ".zip", ".7z"]
    output_dir = output_dir or tempfile.gettempdir()

    if not isinstance(resources, list):
        resources = [resources]

    files = list()

    for arch in resources:
        if any(ext in str(arch).lower() for ext in archive_types):
            try:
                LOGGER.debug("archive=%s", arch)
                file = Path(arch).name

                if file.endswith(".nc"):
                    files.append(Path(output_dir.join(arch)))
                elif file.endswith(".tar"):
                    with tarfile.open(arch, mode="r") as tar:
                        tar.extractall(path=output_dir)
                        files.extend(
                            [str(Path(output_dir).joinpath(f)) for f in tar.getnames()]
                        )
                elif file.endswith(".zip"):
                    with zipfile.ZipFile(arch, mode="r") as zf:
                        zf.extractall(path=output_dir)
                        files.extend(
                            [str(Path(output_dir).joinpath(f)) for f in zf.namelist()]
                        )
                elif file.endswith(".7z"):
                    msg = "7z file extraction is not supported at this time."
                    LOGGER.warning(msg)
                    warnings.warn(msg, UserWarning)
                else:
                    LOGGER.debug('File extension "%s" unknown' % file)
            except Exception as e:
                LOGGER.error(
                    "Failed to extract sub archive {{{}}}: {{{}}}".format(arch, e)
                )
        else:
            LOGGER.warning("No archives found. Continuing...")
            return resources

    return files


def archive_sniffer(
    archives: Union[str, Path, List[Union[str, Path]]],
    working_dir: Optional[Union[str, Path]] = None,
    extensions: Optional[Iterable[str]] = None,
) -> List[Union[str, Path]]:
    """Return a list of locally unarchived files that match the desired extensions.

    Parameters
    ----------
    archives : Union[str, Path, List[Union[str, Path]]]
      archive location or list of archive locations
    working_dir : Union[str, path]
      string or Path to a working location
    extensions : List[str]
      list of accepted extensions

    Returns
    -------
    List[Union[str, Path]]
      List of files with matching accepted extensions
    """
    potential_files = list()

    if not extensions:
        extensions = [".gml", ".shp", ".geojson", ".gpkg", ".json"]

    decompressed_files = generic_extract_archive(archives, output_dir=working_dir)
    for file in decompressed_files:
        if any(ext in Path(file).suffix for ext in extensions):
            potential_files.append(file)
    return potential_files


def crs_sniffer(
    *args: Union[str, Path, Sequence[Union[str, Path]]]
) -> Union[List[Union[str, int]], str, int]:
    """Return the list of CRS found in files.

    Parameters
    ----------
    args : Union[str, Path, Sequence[Union[str, Path]]]
      Path(s) to the file(s) to examine.

    Returns
    -------
    Union[List[str], str]
      Returns either a list of CRSes or a single CRS definition, depending on the number of instances found.
    """
    crs_list = list()
    vectors = (".gml", ".shp", ".geojson", ".gpkg", ".json")
    rasters = (".tif", ".tiff")
    all_files = vectors + rasters

    for file in args:
        found_crs = False
        suffix = Path(file).suffix.lower()
        try:
            if suffix == ".zip":
                file = archive_sniffer(file, extensions=all_files)[0]
                suffix = Path(file).suffix.lower()

            if suffix in vectors:
                if suffix == ".gpkg":
                    if len(fiona.listlayers(file)) > 1:
                        raise NotImplementedError
                with fiona.open(file, "r") as src:
                    found_crs = CRS.from_wkt(src.crs_wkt).to_epsg()
            elif suffix in rasters:
                with rasterio.open(file, "r") as src:
                    found_crs = CRS.from_user_input(src.crs).to_epsg()
            else:
                raise FileNotFoundError("Invalid filename suffix")
        except FileNotFoundError as e:
            msg = f"{e}: Unable to open file {args}"
            LOGGER.warning(msg)
            raise Exception(msg)
        except NotImplementedError as e:
            msg = f"{e}: Multilayer GeoPackages are currently unsupported"
            LOGGER.error(msg)
            raise Exception(msg)
        except RuntimeError:
            pass

        crs_list.append(found_crs)

    if crs_list is None:
        msg = f"No CRS definitions found in {args}."
        raise FileNotFoundError(msg)

    if len(crs_list) == 1:
        if not crs_list[0]:
            msg = f"No CRS definitions found in {args}. Assuming {WGS84}."
            LOGGER.warning(msg)
            warnings.warn(msg, UserWarning)
            return WGS84
        return crs_list[0]
    return crs_list


def raster_datatype_sniffer(file: Union[str, Path]) -> str:
    """Return the type of the raster stored in the file.

    Parameters
    ----------
    file : Union[str, Path]
      Path to file.

    Returns
    -------
    str
      rasterio datatype of array values
    """
    try:
        with rasterio.open(file, "r") as src:
            dtype = src.dtypes[0]
        return dtype
    except rasterio.errors.RasterioError:
        msg = f"Unable to read data type from {file}."
        LOGGER.exception(msg)
        raise ValueError(msg)


def get_bbox(vector: Union[str, Path], all_features: bool = True) -> tuple:
    """Return bounding box of all features or the first feature in file.

    Parameters
    ----------
    vector : str
      Path to file storing vector features.
    all_features : bool
      Return the bounding box for all features. Default: True.

    Returns
    -------
    tuple
      Geographic coordinates of the bounding box (lon0, lat0, lon1, lat1).

    """

    if not all_features:
        with fiona.open(vector, "r") as src:
            for feature in src:
                geom = shape(feature["geometry"])
                return geom.bounds

    with fiona.open(vector, "r") as src:
        return src.bounds



from .base import Raven
from .emulators import get_model


class RavenMultiModel(Raven):
    def __init__(self, models, workdir=None):
        """Create multi-model raven instance.

        Parameters
        ----------
        models : sequence
          Model identifiers ('gr4jcn', 'hmets', 'mohyse', 'hbvec').
        """
        import tempfile

        self._names = models
        self._models = []

        workdir = workdir or tempfile.mkdtemp()
        Raven.__init__(self, workdir)

        for name in models:
            m = get_model(name)(workdir)
            m.model_dir = name
            self._models.append(m)

    def _rename_run_name(self, run_name=None):
        rns = {m.config.rvi.run_name for m in self._models}
        if (run_name is not None) or (len(rns) < len(self._models)):
            for m in self._models:
                rn = run_name or m.config.rvi.run_name
                m.config.rvi.run_name = rn + "_" + m.config.identifier

    def assign(self, key, value):
        """Assign key to all models, unless it's model parameters."""
        # Model parameter case
        if key in self._names:
            m = self._models[self._names.index(key)]
            m.assign("params", value)
        else:
            for m in self._models:
                m.assign(key, value)

    def resume(self, solution=None):
        # TODO: Add support for model dependent solutions.
        for m in self._models:
            m.resume(solution)

    @property
    def _rv_paths(self):
        out = []
        for m in self._models:
            out.extend(m._rv_paths)
        return out

    @_rv_paths.setter
    def _rv_paths(self, value):
        pass

    # @property
    # def rvs(self):
    #     out = []
    #     for m in self._models:
    #         out.extend(m._rv_paths)
    #     return out

    def run(self, ts, overwrite=False, **kwds):
        """Run model.

        Parameters
        ----------
        kwds : dict
          model_name : array
            Parameter array.
        """
        if overwrite:
            self.setup(overwrite)

        self._rename_run_name(kwds.pop("run_name", None))

        p = {}
        for m in self._models:
            p[m.config.identifier] = kwds.pop(m.config.identifier, None)

        procs = []
        for m in self._models:
            # Add params to kwds if passed in run.
            kw = kwds.copy()
            if p[m.config.identifier]:
                kw["params"] = p[m.config.identifier]

            procs.extend(m.run(ts, **kw))

        return procs



from pathlib import Path

import click

from ravenpy.config.commands import GridWeightsCommand
from ravenpy.extractors.routing_product import RoutingProductGridWeightExtractor


@click.command()
@click.argument("input-nc-file", type=click.Path(exists=True))
@click.argument("input-weight-file", type=click.Path(exists=True))
@click.option(
    "-d",
    "--dim-names",
    nargs=2,
    type=click.Tuple([str, str]),
    default=RoutingProductGridWeightExtractor.DIM_NAMES,
    show_default=True,
    help="Ordered dimension names of longitude (x) and latitude (y) in the NetCDF INPUT_NC_FILE.",
)
@click.option(
    "-v",
    "--var-to-aggregate",
    "variables_to_aggregate",  # rename arg to put emphasis on the fact that it's a list
    required=True,
    multiple=True,
    type=str,
    show_default=True,
    help="Variables to aggregate in INPUT_NC_FILE (at least one).",
)
@click.option("--output-nc-file", type=click.Path(), help="")
@click.option("--output-weight-file", type=click.Path(), help="")
def aggregate_forcings_to_hrus(
    input_nc_file,
    input_weight_file,
    output_nc_file,
    output_weight_file,
    dim_names,
    variables_to_aggregate,
):
    """
    Aggregates NetCDF files containing 3-dimensional forcing variables like precipitation and temperature
    over (x,y,time) into 2-dimensional forcings for each of the n HRUs of a specific basin over (n,time).
    The 3-dimensional NetCDF files are usually used in ``:GriddedForcing`` commands in Raven while the 2-dimensional
    ones can be used in ``:StationForcing`` commands. The NetCDF files generated with this function will only
    contain the forcings required to simulate an individual basin and hence file sizes are smaller and Raven
    runtimes can decrease drastically under certain conditions.

    INPUT_NC_FILE: NetCDF file containing 3-dimensional variables that will be aggregated. Either all variables
    will be aggregated or only a subset specified using --var-to-aggregate (e.g., [precip,temp]). The name of
    the spatial dimensions of the NetCDF are assumed to be (lon_dim, lat_dim). Otherwise they will need to be
    specified using --dim-names. The order of the three dimensions for each variable does not matter; the
    function will arrange them as required.

    INPUT_WEIGHT_FILE: A text file containing the grid weights derived using the script "generate-grid-weights"
    for the basin forcings are required and the specified NetCDF file. The content of this file must be formatted
    as a valid ``:GridWeights`` Raven command.

    The script outputs two files:

    (1) Aggregated NetCDF file that can be used in a ``:StationForcing`` command in a Raven config.

    (2) A text file (with the same format as INPUT_WEIGHT_FILE) with the updated grid weights, that a ``:StationForcing``
    command will require.
    """
    # NOTE: This is in order to make sphinx-click happy. Magic. Do not touch.
    import netCDF4 as nc4
    import numpy as np

    gws = GridWeightsCommand.parse(Path(input_weight_file).read_text())

    nHRU = gws.number_hrus
    # nCells = gws.number_grid_cells
    weights_data = gws.data

    # read NetCDF
    nc_in = nc4.Dataset(input_nc_file, "r")

    # length of dimensions
    nlon = nc_in.dimensions[dim_names[0]].size
    # nlat = nc_in.dimensions[dim_names[1]].size
    ntime = nc_in.dimensions["time"].size

    # convert weights (cell_id) required by Raven into (lon_id, lat_id)
    # cell_id = ilat * nlon + ilon
    # ---> ilon = cell_id %  nlon
    # ---> ilat = cell_id // nlon
    weights_data_lon_lat_ids = []
    for iweights_data in weights_data:
        cell_id = iweights_data[1]
        ilon = cell_id % nlon
        ilat = cell_id // nlon
        weights_data_lon_lat_ids.append(
            tuple([iweights_data[0], ilon, ilat, iweights_data[2]])
        )
    weights_data_lon_lat_ids = np.asarray(weights_data_lon_lat_ids)

    # create new NetCDF that will contain aggregated data of listed variables

    if not output_nc_file:
        input_nc_file_path = Path(input_nc_file)
        output_nc_file_path = (
            input_nc_file_path.parent / f"{input_nc_file_path.stem}_aggregated.nc"
        )
    else:
        output_nc_file_path = Path(output_nc_file)

    nc_out = nc4.Dataset(output_nc_file_path, "w")
    _ = nc_out.createDimension("time", ntime)
    _ = nc_out.createDimension("nHRU", nHRU)

    # copy all global attributes over
    nc_out.setncatts(nc_in.__dict__)

    # create all variables in output NC (incl. time) and copy over all attributes
    for name, variable in nc_in.variables.items():
        if name in variables_to_aggregate + ("time",):

            if name != "time":
                dims = ["time", "nHRU"]
            else:
                dims = ["time"]
            _ = nc_out.createVariable(name, variable.datatype, dims, zlib=True)

            # copy variable attributes all at once via dictionary
            nc_out[name].setncatts(nc_in[name].__dict__)

            # copy over time variable data
            if name == "time":
                nc_out[name][:] = nc_in[name][:]

    for variable_to_aggregate in variables_to_aggregate:

        # read 3D variable
        input_var = nc_in.variables[variable_to_aggregate]

        # read 2D variable (to be able to write to it later)
        output_var = nc_out.variables[variable_to_aggregate]

        # is variable 3D ?
        ndims = len(input_var.dimensions)
        if ndims != 3:
            raise ValueError("Input variable to aggregate needs to be 3D")

        # what is the order of dimensions?
        idx_lon_dim = input_var.dimensions.index(dim_names[0])
        idx_lat_dim = input_var.dimensions.index(dim_names[1])
        # idx_time_dim = input_var.dimensions.index("time")

        # read in data for bounding box (is faster than reading then every single cell individually)
        # --> this takes most time for large NetCDFs
        min_lon = int(np.min(weights_data_lon_lat_ids[:, 1]))
        max_lon = int(np.max(weights_data_lon_lat_ids[:, 1]))
        min_lat = int(np.min(weights_data_lon_lat_ids[:, 2]))
        max_lat = int(np.max(weights_data_lon_lat_ids[:, 2]))

        idx_input = [slice(0, ntime, 1), slice(0, ntime, 1), slice(0, ntime, 1)]
        idx_input[idx_lon_dim] = slice(min_lon, max_lon + 1, 1)
        idx_input[idx_lat_dim] = slice(min_lat, max_lat + 1, 1)
        idx_input = tuple(idx_input)

        # print(weights_data_lon_lat_ids[:, 1])
        # print(min_lon, max_lon)
        # print(min_lat, max_lat)
        # print(idx_input)

        input_var_bb = input_var[idx_input]

        # list of HRU IDs
        hrus = np.unique(weights_data_lon_lat_ids[:, 0])

        if len(hrus) != nHRU:
            # should really never happen
            raise ValueError(
                "Number of weights found in grid weights list is not matching the number indicated there by nHRUs"
            )

        # do actual aggregation
        agg_var = np.zeros([ntime, nHRU])
        new_weights = []
        for ihru, hru in enumerate(hrus):

            hru = int(hru)

            # filter all weights for current HRU
            idx = np.where(weights_data_lon_lat_ids[:, 0] == hru)[0]

            # create an array that contains weights for each time step
            # --> weights need to be rescaled if NODATA values appear
            # --> transpose such that we can broadcast during rescaling later
            # --> shape = (ncells,ntime) with ncells = len(idx)
            weights_nodata = np.transpose(
                np.tile(weights_data_lon_lat_ids[idx, 3], (ntime, 1))
            )

            # go through all time steps and zero out weights where grid cell is NODATA
            for iii, ii in enumerate(idx):

                # bring idx for input_var in appropriate order
                idx_input = [slice(0, ntime, 1), slice(0, ntime, 1), slice(0, ntime, 1)]
                idx_input[idx_lon_dim] = int(weights_data_lon_lat_ids[ii, 1]) - min_lon
                idx_input[idx_lat_dim] = int(weights_data_lon_lat_ids[ii, 2]) - min_lat
                idx_input = tuple(idx_input)

                weights_nodata[iii] = np.where(
                    input_var_bb[idx_input].mask, 0.0, weights_nodata[iii]
                )

            # rescale
            # --> columns where all grid cells are nan might cause a warning:
            #     "RuntimeWarning: invalid value encountered in true_divide"
            old_settings = np.seterr()
            np.seterr(invalid="ignore")
            # "nan_to_num" automatically converts NaN's to 0
            weights_nodata = np.nan_to_num(
                weights_nodata / np.sum(weights_nodata, axis=0)
            )
            np.seterr(**old_settings)  # reset to default

            # derive aggregate
            for iii, ii in enumerate(idx):

                # bring idx for input_var in appropriate order
                idx_input = [slice(0, ntime, 1), slice(0, ntime, 1), slice(0, ntime, 1)]
                idx_input[idx_lon_dim] = int(weights_data_lon_lat_ids[ii, 1]) - min_lon
                idx_input[idx_lat_dim] = int(weights_data_lon_lat_ids[ii, 2]) - min_lat
                idx_input = tuple(idx_input)

                # this does not work when NODATA values occur
                # agg_var[:, ihru] += (
                #     input_var_bb[idx_input] * weights_data_lon_lat_ids[ii, 3]
                # )

                # this makes sure to handle NODATA cells properly
                agg_var[:, ihru] += input_var_bb[idx_input].data * weights_nodata[iii]

            # if all grid cells have weight 0.0 (i.e., all invalid) set value to NODATA
            idx_t_all_cells_nodata = np.where(
                np.sum(weights_nodata[:, :], axis=0) == 0.0
            )
            if len(idx_t_all_cells_nodata[0]) > 0:
                # only if such cells are found
                # otherwise "input_var.missing_value" might not even exist
                agg_var[idx_t_all_cells_nodata, ihru] = input_var.missing_value

            # create new weights: now each HRU is exactly one "grid-cell"
            new_weights.append(tuple([hru, ihru, 1.0]))

        # write 2D variable
        output_var[:] = agg_var

    nc_out.close()

    # the return should actually look exactly like the return of "RoutingProductGridWeightImporter"
    # not sure how to do that

    gws_new = GridWeightsCommand(
        number_hrus=nHRU,
        number_grid_cells=nHRU,
        data=new_weights,
    )

    if not output_weight_file:
        input_weight_file_path = Path(input_weight_file)
        output_weight_file_path = (
            input_weight_file_path.parent
            / f"{input_weight_file_path.stem}_aggregated.rvt"
        )
    else:
        output_weight_file_path = Path(output_weight_file)

    output_weight_file_path.write_text(gws_new.to_rv() + "\n")

    click.echo(f"Created {output_nc_file_path}")
    click.echo(f"Created {output_weight_file_path}")



"""
Tools for reading outputs and writing inputs for the Raven executable.
"""

from collections import OrderedDict


# TODO: Implement section parser
def parse_configuration(fn):
    """Parse Raven configuration file.

    Returns a dictionary keyed by parameter name."""
    import re

    main_param = re.compile(r"^:(\w+)\s+([^#]*)")
    # sub_param = re.compile(r"^  :(\w+)\s+([^#]*)")
    out = OrderedDict()
    # cat = None
    with open(str(fn)) as f:
        for line in f.readlines():
            match = main_param.search(line)
            if not match:
                continue

            key, value = match.groups()
            if value:
                values = value.split()
                out[key] = values[0] if len(values) == 1 else values
            else:
                if "List" in key:
                    pass
                elif "Classes" in key:
                    pass
                elif "Profiles" in key:
                    pass
                else:
                    out[key] = True

    return out



"""
Tools for performing geospatial translations and transformations.
"""

import collections
import json
import logging
from pathlib import Path
from typing import List, Union

from . import gis_import_error_message

try:
    import fiona
    import rasterio
    import rasterio.mask
    import rasterio.vrt
    import rasterio.warp
    from pyproj import CRS
    from shapely.geometry import (
        GeometryCollection,
        MultiPolygon,
        Polygon,
        mapping,
        shape,
    )
    from shapely.ops import transform
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

RASTERIO_TIFF_COMPRESSION = "lzw"
LOGGER = logging.getLogger("RavenPy")
WGS84 = 4326


def geom_transform(
    geom: Union[GeometryCollection, shape],
    source_crs: Union[str, int, CRS] = WGS84,
    target_crs: Union[str, int, CRS] = None,
) -> GeometryCollection:
    """Change the projection of a geometry.

    Assuming a geometry's coordinates are in a `source_crs`, compute the new coordinates under the `target_crs`.

    Parameters
    ----------
    geom : Union[GeometryCollection, shape]
      Source geometry.
    source_crs : Union[str, int, CRS]
      Projection identifier (proj4) for the source geometry, e.g. '+proj=longlat +datum=WGS84 +no_defs'.
    target_crs : Union[str, int, CRS]
      Projection identifier (proj4) for the target geometry.

    Returns
    -------
    GeometryCollection
      Reprojected geometry.
    """
    try:
        from functools import partial

        from pyproj import Transformer  # noqa

        if isinstance(source_crs, int or str):
            source = CRS.from_epsg(source_crs)
        else:
            source = source_crs

        if isinstance(target_crs, int or str):
            target = CRS.from_epsg(target_crs)
        else:
            target = target_crs

        transform_func = Transformer.from_crs(source, target, always_xy=True)
        reprojected = transform(transform_func.transform, geom)

        return reprojected
    except Exception as err:
        msg = f"{err}: Failed to reproject geometry"
        LOGGER.error(msg)
        raise Exception(msg)


def generic_raster_clip(
    raster: Union[str, Path],
    output: Union[str, Path],
    geometry: Union[Polygon, MultiPolygon, List[Union[Polygon, MultiPolygon]]],
    touches: bool = False,
    fill_with_nodata: bool = True,
    padded: bool = True,
    raster_compression: str = RASTERIO_TIFF_COMPRESSION,
) -> None:
    """
    Crop a raster file to a given geometry.

    Parameters
    ----------
    raster : Union[str, Path]
      Path to input raster.
    output : Union[str, Path]
      Path to output raster.
    geometry : Union[Polygon, MultiPolygon, List[Union[Polygon, MultiPolygon]]
      Geometry defining the region to crop.
    touches : bool
      Whether or not to include cells that intersect the geometry. Default: True.
    fill_with_nodata: bool
      Whether or not to keep pixel values for regions outside of shape or set as nodata. Default: True.
    padded: bool
      Whether or not to add a half-pixel buffer to shape before masking
    raster_compression : str
      Level of data compression. Default: 'lzw'.

    Returns
    -------
    None
    """
    if not isinstance(geometry, collections.abc.Iterable):
        geometry = [geometry]

    with rasterio.open(raster, "r") as src:
        mask_image, mask_affine = rasterio.mask.mask(
            src,
            geometry,
            crop=True,
            pad=padded,
            all_touched=touches,
            filled=fill_with_nodata,
        )
        mask_meta = src.meta.copy()
        mask_meta.update(
            {
                "driver": "GTiff",
                "height": mask_image.shape[1],
                "width": mask_image.shape[2],
                "transform": mask_affine,
                "compress": raster_compression,
            }
        )

        # Write the new masked image
        with rasterio.open(output, "w", **mask_meta) as dst:
            dst.write(mask_image)


def generic_raster_warp(
    raster: Union[str, Path],
    output: Union[str, Path],
    target_crs: Union[str, dict, CRS],
    raster_compression: str = RASTERIO_TIFF_COMPRESSION,
) -> None:
    """
    Reproject a raster file.

    Parameters
    ----------
    raster : Union[str, Path]
      Path to input raster.
    output : Union[str, Path]
      Path to output raster.
    target_crs : str or dict
      Target projection identifier.
    raster_compression: str
      Level of data compression. Default: 'lzw'.

    Returns
    -------
    None
    """
    with rasterio.open(raster, "r") as src:
        # Reproject raster using WarpedVRT class
        with rasterio.vrt.WarpedVRT(src, crs=target_crs) as vrt:
            # Calculate grid properties based on projection
            affine, width, height = rasterio.warp.calculate_default_transform(
                src.crs, target_crs, src.width, src.height, *src.bounds
            )

            # Copy relevant metadata from parent raster
            metadata = src.meta.copy()
            metadata.update(
                {
                    "driver": "GTiff",
                    "height": height,
                    "width": width,
                    "transform": affine,
                    "crs": target_crs,
                    "compress": raster_compression,
                }
            )
            data = vrt.read()

            with rasterio.open(output, "w", **metadata) as dst:
                dst.write(data)


def generic_vector_reproject(
    vector: Union[str, Path],
    projected: Union[str, Path],
    source_crs: Union[str, CRS] = WGS84,
    target_crs: Union[str, CRS] = None,
) -> None:
    """Reproject all features and layers within a vector file and return a GeoJSON

    Parameters
    ----------
    vector : Union[str, Path]
      Path to a file containing a valid vector layer.
    projected: Union[str, Path]
      Path to a file to be written.
    source_crs : Union[str, dict, CRS]
      Projection identifier (proj4) for the source geometry, Default: '+proj=longlat +datum=WGS84 +no_defs'.
    target_crs : Union[str, dict, CRS]
      Projection identifier (proj4) for the target geometry.

    Returns
    -------
    None
    """

    if target_crs is None:
        raise ValueError("No target CRS is defined.")

    output = {"type": "FeatureCollection", "features": list()}

    if isinstance(vector, Path):
        vector = vector.as_posix()

    for i, layer_name in enumerate(fiona.listlayers(vector)):
        with fiona.open(vector, "r", layer=i) as src:
            with open(projected, "w") as sink:
                for feature in src:
                    # Perform vector reprojection using Shapely on each feature
                    try:
                        geom = shape(feature["geometry"])
                        transformed = geom_transform(geom, source_crs, target_crs)
                        feature["geometry"] = mapping(transformed)
                        output["features"].append(feature)
                    except Exception as err:
                        LOGGER.exception(
                            "{}: Unable to reproject feature {}".format(err, feature)
                        )
                        raise

                sink.write(f"{json.dumps(output)}")



import collections
import datetime as dt
from abc import ABC, abstractmethod
from dataclasses import is_dataclass, replace
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import Dict, Optional, Tuple

import cf_xarray
import cftime
import numpy as np
import xarray as xr

from ravenpy.config.commands import (
    BasinIndexCommand,
    BasinStateVariablesCommand,
    ChannelProfileCommand,
    DataCommand,
    GaugeCommand,
    GriddedForcingCommand,
    GridWeightsCommand,
    HRUsCommand,
    HRUState,
    HRUStateVariableTableCommand,
    LandUseClassesCommand,
    ObservationDataCommand,
    SoilClassesCommand,
    SoilProfilesCommand,
    StationForcingCommand,
    SubBasinGroupCommand,
    SubBasinsCommand,
    VegetationClassesCommand,
)


class RV(ABC):
    def __init__(self, config, **kwds):
        # Each RV has a reference to their parent object in order to access sibling RVs.
        self._config = config
        self.is_ostrich_tmpl = False
        self.content = None

    def update(self, key, value):
        if hasattr(self, key):
            # Special case: the user might be trying to update `params` or `derived_params`,
            # (which are dataclasses) by supplying a list of values (either a list, tuple of
            # numpy array); if that's the case, cast those values into a new instance of the
            # corresponding dataclass.
            attr = getattr(self, key)
            if is_dataclass(attr) and isinstance(value, (list, tuple, np.ndarray)):
                value = attr.__class__(*value)
            setattr(self, key, value)
            return True
        return False

    def get_extra_attributes(self, d):
        """
        Not sure about this: for the moment I use it only for certain params that must
        be injected in the RVH by the MOHYSE emulator. The idea is to complete the `d`
        dict used in the `to_rv` method for the template with extra attributes that have
        been added in the emulator.
        """
        e = {}
        for k, v in self.__dict__.items():
            if k not in d:
                e[k] = v
        return e

    def set_tmpl(self, tmpl, is_ostrich=False):
        self.tmpl = tmpl
        self.is_ostrich_tmpl = is_ostrich

    @property
    @abstractmethod
    def tmpl(self):
        pass

    @abstractmethod
    def to_rv(self) -> str:
        pass


#########
# R V C #
#########


class RVC(RV):

    tmpl = """
    {hru_states}

    {basin_states}
    """

    def __init__(self, config):
        super().__init__(config)
        self.hru_states: Dict[int, HRUState] = {}
        self.basin_states: Dict[int, BasinIndexCommand] = {}

    def reset(self, **kwargs):
        self.hru_states = {}
        self.basin_states = {}

    def set_hru_state(self, hru_state: HRUState):
        self.hru_states[hru_state.index] = hru_state

    def set_basin_state(self, basin_state: BasinIndexCommand):
        self.basin_states[basin_state.index] = basin_state

    @classmethod
    def create_solution(cls, solution_str):
        rvc = RVC(None)
        rvc.parse_solution(solution_str)
        return rvc

    def parse_solution(self, solution_str):
        self.hru_states = HRUStateVariableTableCommand.parse(solution_str).hru_states
        self.basin_states = BasinStateVariablesCommand.parse(solution_str).basin_states

    def to_rv(self):
        d = {
            "hru_states": HRUStateVariableTableCommand(self.hru_states),
            "basin_states": BasinStateVariablesCommand(self.basin_states),
        }

        d.update(self.get_extra_attributes(d))

        return dedent(self.tmpl).format(**d)


#########
# R V H #
#########


class RVH(RV):

    tmpl = """
    {subbasins}

    {hrus}

    {land_subbasin_group}

    {land_subbasin_property_multiplier}

    {lake_subbasin_group}

    {lake_subbasin_property_multiplier}

    {reservoirs}
    """

    def __init__(self, config):
        super().__init__(config)
        self.hrus = ()
        self.subbasins = ()
        self.land_subbasin_ids = ()
        self.land_subbasin_property_multiplier = None
        self.lake_subbasin_ids = ()
        self.lake_subbasin_property_multiplier = None
        self.reservoirs = ()

    def to_rv(self):
        d = {
            "subbasins": SubBasinsCommand(self.subbasins),
            "hrus": HRUsCommand(self.hrus),
            "land_subbasin_group": SubBasinGroupCommand("Land", self.land_subbasin_ids),
            "land_subbasin_property_multiplier": self.land_subbasin_property_multiplier
            or "",
            "lake_subbasin_group": SubBasinGroupCommand(
                "Lakes", self.lake_subbasin_ids
            ),
            "lake_subbasin_property_multiplier": self.lake_subbasin_property_multiplier
            or "",
            "reservoirs": "\n\n".join(map(str, self.reservoirs)),
        }

        d.update(self.get_extra_attributes(d))

        return dedent(self.tmpl).format(**d)


#########
# R V I #
#########


class RVI(RV):

    _pre_tmpl = """
    :Calendar              {calendar}
    :RunName               {run_name}-{run_index}
    :StartDate             {start_date}
    :EndDate               {end_date}
    :TimeStep              {time_step}
    :Method                ORDERED_SERIES
    """

    tmpl = """
    """

    # This part must be at the end of the file in particular because `:EvaluationMetrics` must
    # come after `:SoilModel`
    _post_tmpl = """
    :EvaluationMetrics     {evaluation_metrics}
    :WriteNetcdfFormat     yes
    #:WriteForcingFunctions
    :SilentMode
    :PavicsMode
    {suppress_output}

    :NetCDFAttribute title Simulated river discharge
    :NetCDFAttribute history Created on {now} by Raven
    :NetCDFAttribute references  Craig, J.R., and the Raven Development Team, Raven user's and developer's manual (Version 2.8), URL: http://raven.uwaterloo.ca/ (2018).
    :NetCDFAttribute comment Raven Hydrological Framework version {raven_version}
    :NetCDFAttribute model_id {identifier}
    :NetCDFAttribute time_frequency day
    :NetCDFAttribute time_coverage_start {start_date}
    :NetCDFAttribute time_coverage_end {end_date}
    """

    class EvaporationOptions(Enum):
        PET_CONSTANT = "PET_CONSTANT"
        PET_PENMAN_MONTEITH = "PET_PENMAN_MONTEITH"
        PET_PENMAN_COMBINATION = "PET_PENMAN_COMBINATION"
        PET_PRIESTLEY_TAYLOR = "PET_PRIESTLEY_TAYLOR"
        PET_HARGREAVES = "PET_HARGREAVES"
        PET_HARGREAVES_1985 = "PET_HARGREAVES_1985"
        PET_FROMMONTHLY = "PET_FROMMONTHLY"
        PET_DATA = "PET_DATA"
        PET_HAMON_1961 = "PET_HAMON_1961"
        PET_TURC_1961 = "PET_TURC_1961"
        PET_MAKKINK_1957 = "PET_MAKKINK_1957"
        PET_MONTHLY_FACTOR = "PET_MONTHLY_FACTOR"
        PET_MOHYSE = "PET_MOHYSE"
        PET_OUDIN = "PET_OUDIN"

    class CalendarOptions(Enum):
        PROLEPTIC_GREGORIAN = "PROLEPTIC_GREGORIAN"
        JULIAN = "JULIAN"
        GREGORIAN = "GREGORIAN"
        STANDARD = "STANDARD"
        NOLEAP = "NOLEAP"
        _365_DAY = "365_DAY"
        ALL_LEAP = "ALL_LEAP"
        _366_DAY = "366_DAY"

    class EvaluationMetrics(Enum):
        NASH_SUTCLIFFE = "NASH_SUTCLIFFE"
        LOG_NASH = "LOG_NASH"
        RMSE = "RMSE"
        PCT_BIAS = "PCT_BIAS"
        ABSERR = "ABSERR"
        ABSMAX = "ABSMAX"
        PDIFF = "PDIFF"
        TMVOL = "TMVOL"
        RCOEFF = "RCOEFF"
        NSC = "NSC"
        KLING_GUPTA = "KLING_GUPTA"

    class RainSnowFractionOptions(Enum):
        DATA = "RAINSNOW_DATA"
        DINGMAN = "RAINSNOW_DINGMAN"
        UBC = "RAINSNOW_UBC"
        HBV = "RAINSNOW_HBV"
        HARDER = "RAINSNOW_HARDER"
        HSPF = "RAINSNOW_HSPF"

    class RoutingOptions(Enum):
        DIFFUSIVE_WAVE = "ROUTE_DIFFUSIVE_WAVE"
        HYDROLOGIC = "ROUTE_HYDROLOGIC"
        NONE = "ROUTE_NONE"
        STORAGE_COEFF = "ROUTE_STORAGE_COEFF"
        PLUG_FLOW = "ROUTE_PLUG_FLOW"
        MUSKINGUM = "MUSKINGUM"

    def __init__(self, config):
        super().__init__(config)

        # These are attributes that can be modified/set directly
        self.run_name = "run"
        self.run_index = 0
        self.raven_version = "3.X.X"
        self.time_step = 1.0

        # These correspond to properties whose setters will pass their value through
        # an Enum cast (triggering a ValueError at runtime for unknown values) and
        # getters will be used when rendering the template in the `to_rv` method
        self._calendar = RVI.CalendarOptions.STANDARD
        self._routing = RVI.RoutingOptions.NONE
        self._start_date = None
        self._end_date = None
        self._rain_snow_fraction = RVI.RainSnowFractionOptions.DATA
        self._evaporation = None
        self._ow_evaporation = None
        self._duration = 1
        self._evaluation_metrics = [
            RVI.EvaluationMetrics.NASH_SUTCLIFFE,
            RVI.EvaluationMetrics.RMSE,
        ]
        self._suppress_output = False

    @property
    def start_date(self):
        return self._start_date

    @start_date.setter
    def start_date(self, x):
        if isinstance(x, dt.datetime):
            self._start_date = self._dt2cf(x)
        else:
            raise ValueError("Must be datetime")

        if x == dt.datetime(1, 1, 1):
            return

        if self._duration is None:
            self._update_duration()
        else:
            self._update_end_date()

    @property
    def end_date(self):
        return self._end_date

    @end_date.setter
    def end_date(self, x):
        if isinstance(x, dt.datetime):
            self._end_date = self._dt2cf(x)
        else:
            raise ValueError("Must be datetime")

        if x != dt.datetime(1, 1, 1):
            self._update_duration()

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, x):
        if isinstance(x, int):
            if x > 0:
                self._duration = x
        else:
            raise ValueError("Must be int")

        if x > 0:
            self._update_end_date()

    @property
    def evaluation_metrics(self):
        if self._evaluation_metrics:
            return ",".join(m.value for m in self._evaluation_metrics)
        return None

    @evaluation_metrics.setter
    def evaluation_metrics(self, values):
        if not isinstance(values, (list, set, tuple)):
            values = [values]
        ms = []
        for v in values:
            v = v.upper() if isinstance(v, str) else v.value
            ms.append(RVI.EvaluationMetrics(v))
        self._evaluation_metrics = ms

    def _update_duration(self):
        if self.end_date is not None and self.start_date is not None:
            self._duration = (self.end_date - self.start_date).days

    def _update_end_date(self):
        if self.start_date is not None and self.duration is not None:
            self._end_date = self.start_date + dt.timedelta(days=self.duration)

    @property
    def suppress_output(self):
        tag = ":SuppressOutput\n:DontWriteWatershedStorage"
        return tag if self._suppress_output else ""

    @suppress_output.setter
    def suppress_output(self, value):
        if not isinstance(value, bool):
            raise ValueError
        self._suppress_output = value

    @property
    def routing(self):
        return self._routing.value

    @routing.setter
    def routing(self, value):
        v = value.upper() if isinstance(value, str) else value.value
        self._routing = RVI.RoutingOptions(v)

    @property
    def rain_snow_fraction(self):
        """Rain snow partitioning."""
        return self._rain_snow_fraction.value

    @rain_snow_fraction.setter
    def rain_snow_fraction(self, value):
        v = value.upper() if isinstance(value, str) else value.value
        self._rain_snow_fraction = RVI.RainSnowFractionOptions(v)

    @property
    def evaporation(self):
        """Evaporation scheme"""
        return self._evaporation.value if self._evaporation else None

    @evaporation.setter
    def evaporation(self, value):
        v = value.upper() if isinstance(value, str) else value.value
        self._evaporation = RVI.EvaporationOptions(v)

    @property
    def ow_evaporation(self):
        """Open-water evaporation scheme"""
        return self._ow_evaporation.value if self._ow_evaporation else None

    @ow_evaporation.setter
    def ow_evaporation(self, value):
        v = value.upper() if isinstance(value, str) else value.value
        self._ow_evaporation = RVI.EvaporationOptions(v)

    @property
    def calendar(self):
        """Calendar"""
        return self._calendar.value

    @calendar.setter
    def calendar(self, value):
        v = value.upper() if isinstance(value, str) else value.value
        self._calendar = RVI.CalendarOptions(v)

    def _dt2cf(self, date):
        """Convert datetime to cftime datetime."""
        return cftime._cftime.DATE_TYPES[self.calendar.lower()](*date.timetuple()[:6])

    def to_rv(self):

        # Attributes (not starting with "_")
        a = list(filter(lambda x: not x.startswith("_"), self.__dict__))

        # Properties (computing values corresponding to attributes starting with "_')
        p = list(
            filter(
                lambda x: isinstance(getattr(self.__class__, x, None), property),
                dir(self),
            )
        )

        d = {attr: getattr(self, attr) for attr in a + p}

        d["identifier"] = self._config.identifier
        d["now"] = dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        t = dedent(self._pre_tmpl) + dedent(self.tmpl) + dedent(self._post_tmpl)
        return t.format(**d)


##########
# R V P  #
##########


class RVP(RV):

    # This is expected to be defined by the emulators.
    tmpl = """
    """

    def __init__(self, config):
        super().__init__(config)

        # Model specific params and derived params
        self.params = None
        self.derived_params = None

        self.soil_classes: Tuple[SoilClassesCommand.Record, ...] = ()
        self.soil_profiles: Tuple[SoilProfilesCommand.Record, ...] = ()
        self.vegetation_classes: Tuple[VegetationClassesCommand.Record, ...] = ()
        self.land_use_classes: Tuple[LandUseClassesCommand.Record, ...] = ()
        self.channel_profiles: Tuple[ChannelProfileCommand, ...] = ()
        self.avg_annual_runoff: Optional[float] = None

    def to_rv(self):
        d = {
            "params": self.params,
            "derived_params": self.derived_params,
            "soil_classes": SoilClassesCommand(self.soil_classes),
            "soil_profiles": SoilProfilesCommand(self.soil_profiles),
            "vegetation_classes": VegetationClassesCommand(self.vegetation_classes),
            "land_use_classes": LandUseClassesCommand(self.land_use_classes),
            "channel_profiles": "\n\n".join(map(str, self.channel_profiles)),
            "avg_annual_runoff": f":AvgAnnualRunoff {self.avg_annual_runoff}"
            if self.avg_annual_runoff
            else "",
        }
        return dedent(self.tmpl).format(**d)


#########
# R V T #
#########


class RVT(RV):

    tmpl = """
    {gauge}

    {forcing_list}

    {observed_data}
    """

    # Keys are standard names
    NC_VARS = {
        "tasmin": {"raven": "TEMP_MIN", "alts": ["tmin"]},
        "tasmax": {"raven": "TEMP_MAX", "alts": ["tmax"]},
        "tas": {"raven": "TEMP_AVE", "alts": ["t2m"]},
        "rainfall": {"raven": "RAINFALL", "alts": ["rain"]},
        "pr": {
            "raven": "PRECIP",
            "alts": ["precip", "prec", "precipitation", "tp"],
        },
        "prsn": {
            "raven": "SNOWFALL",
            "alts": ["snow", "snowfall", "solid_precip"],
        },
        "evspsbl": {"raven": "PET", "alts": ["pet", "evap", "evapotranspiration"]},
        "water_volume_transport_in_river_channel": {
            "raven": "HYDROGRAPH",
            "alts": [
                "qobs",
                "discharge",
                "streamflow",
                "dis",
            ],
        },
    }

    def __init__(self, config):
        super().__init__(config)

        self._var_cmds = {k: {} for k in RVT.NC_VARS.keys()}

        self.nc_index = 0
        self.grid_weights = None
        self.rain_correction = None
        self.snow_correction = None
        self.monthly_ave_evaporation = None
        self.monthly_ave_temperature = None

        self._nc_latitude = []
        self._nc_longitude = []
        self._nc_elevation = []
        self._number_grid_cells = 0

    def add_nc_variable(self, **kwargs):
        std_name = kwargs.get("name", kwargs["var_name_nc"])
        is_obs_var = kwargs.pop("is_observation", False)
        if len(kwargs["dim_names_nc"]) == 1:
            if std_name == "water_volume_transport_in_river_channel" or is_obs_var:
                cmd = ObservationDataCommand(**kwargs)
            else:
                cmd = DataCommand(**kwargs)
        elif len(kwargs["dim_names_nc"]) == 2:
            if std_name == "water_volume_transport_in_river_channel" or is_obs_var:
                cmd = ObservationDataCommand(**kwargs)
            else:
                cmd = StationForcingCommand(**kwargs)
        else:
            cmd = GriddedForcingCommand(**kwargs)

        if isinstance(self._var_cmds.get(std_name, None), dict):
            self._var_cmds[std_name] = replace(cmd, **self._var_cmds[std_name])
        else:
            self._var_cmds[std_name] = cmd

    def configure_from_nc_data(self, fns):

        # Important note: if the object at key `k` is a dict (as opposed to a `Command`),
        # don't reset it because it contains initial user-defined config (for the future Command
        # object at that particular key)
        for std_name in RVT.NC_VARS:
            v = self._var_cmds[std_name]
            if not isinstance(v, dict):
                self._var_cmds[std_name] = {}

        for fn in fns:
            with xr.open_dataset(fn) as ds:
                try:
                    self.nc_latitude = ds.cf["latitude"]
                    self.nc_longitude = ds.cf["longitude"]
                    self.nc_elevation = ds.cf["vertical"]
                except KeyError:
                    # Will try to compute values later from first HRU (in self.to_rv)
                    pass

                # Check if any alternate variable name is in the file.
                for std_name in RVT.NC_VARS:
                    for var_name in [std_name] + RVT.NC_VARS[std_name]["alts"]:
                        if var_name not in ds.data_vars:
                            continue
                        nc_var = ds[var_name]
                        self.add_nc_variable(
                            name=std_name,
                            file_name_nc=fn,
                            data_type=RVT.NC_VARS[std_name]["raven"],
                            var_name_nc=var_name,
                            dim_names_nc=nc_var.dims,
                            units=nc_var.attrs.get("units"),
                        )
                        self._number_grid_cells = int(nc_var.size / len(ds["time"]))
                        break

    def update(self, key, value):
        if key in self._var_cmds:
            self._var_cmds[key].update(value)
            return True
        elif key == "nc_index":
            self.nc_index = value
            return True
        return False

    def to_rv(self):
        """
        IMPORTANT NOTE: as this method is called at the last moment in the model lifecycle,
        we can take the occasion to inject in the data structure some values that are guaranteed
        to be there (for instance we can assume that the RVH data is fully specified).
        """

        d = {
            "gauge": "",
            "forcing_list": "",
            "observed_data": "",
        }

        use_gauge = any(type(cmd) is DataCommand for cmd in self._var_cmds.values())
        if use_gauge:
            data = []
            for var, cmd in self._var_cmds.items():
                if cmd and not isinstance(cmd, ObservationDataCommand):
                    data.append(cmd)
            lat = (
                self._nc_latitude[self.nc_index]
                if self._nc_latitude
                else self._config.rvh.hrus[0].latitude
            )
            lon = (
                self._nc_longitude[self.nc_index]
                if self._nc_longitude
                else self._config.rvh.hrus[0].longitude
            )
            elev = (
                self._nc_elevation[self.nc_index]
                if self._nc_elevation
                else self._config.rvh.hrus[0].elevation
            )

            d["gauge"] = GaugeCommand(
                latitude=lat,
                longitude=lon,
                elevation=elev,
                rain_correction=self.rain_correction,
                snow_correction=self.snow_correction,
                monthly_ave_evaporation=self.monthly_ave_evaporation,
                monthly_ave_temperature=self.monthly_ave_temperature,
                data=data,
            )
        else:
            # Construct default grid weights applying equally to all HRUs
            data = [(hru.hru_id, self.nc_index, 1.0) for hru in self._config.rvh.hrus]
            gw = self.grid_weights or GridWeightsCommand(
                number_hrus=len(data),
                number_grid_cells=self._number_grid_cells,
                data=data,
            )
            cmds = []
            for var, cmd in self._var_cmds.items():
                if cmd and not isinstance(cmd, ObservationDataCommand):
                    # TODO: implement a RedirectToFile mechanism to avoid inlining the grid weights
                    # multiple times as we do here
                    if len(cmd.grid_weights.data) == 1:
                        cmd.grid_weights = gw
                    cmds.append(cmd)
            d["forcing_list"] = "\n".join(map(str, cmds))

        # QUESTION: is it possible to have (and if yes should we support) more than 1
        # observation variable? For now we don't.
        for cmd in self._var_cmds.values():
            if isinstance(cmd, ObservationDataCommand):
                # Search for the gauged SB, not sure what should happen when there are
                # more than one (should it be even supported?)
                for sb in self._config.rvh.subbasins:
                    if sb.gauged:
                        cmd.subbasin_id = sb.subbasin_id
                        break
                else:
                    raise Exception(
                        "Could not find an outlet subbasin for observation data"
                    )
                # Set the :StationxIdx (which starts at 1)
                cmd.index = self.nc_index + 1
                d["observed_data"] = cmd
                break

        return dedent(self.tmpl).format(**d)


#########
# O s t #
#########


class OST(RV):

    tmpl = """
    """

    def __init__(self, config):
        super().__init__(config)

        self._max_iterations = None
        self._random_seed = None
        self.lowerBounds = None
        self.upperBounds = None
        self.algorithm = None
        # If there's an OstRandomNumbers.txt file this is its path
        self.random_numbers_path = None

    @property
    def max_iterations(self):
        return self._max_iterations

    @max_iterations.setter
    def max_iterations(self, x):
        if x < 1:
            raise ValueError("Max iteration should be a positive integer: {}".format(x))
        else:
            self._max_iterations = x

    @property
    def random_seed(self):
        if self._random_seed is not None:
            return "RandomSeed {}".format(self._random_seed)
        return ""

    @random_seed.setter
    def random_seed(self, value):
        if value >= 0:
            self._random_seed = value
        else:
            self._random_seed = None

    @property
    def identifier(self):
        return self._config.identifier

    def to_rv(self):
        # Get those from RVI (there's probably a better way to do this!)
        self.run_name = self._config.rvi.run_name
        self.run_index = self._config.rvi.run_index

        # Attributes
        a = list(filter(lambda x: not x.startswith("_"), self.__dict__))

        # Properties
        p = list(
            filter(
                lambda x: isinstance(getattr(self.__class__, x, None), property),
                dir(self),
            )
        )

        d = {attr: getattr(self, attr) for attr in a + p}

        return dedent(self.tmpl).format(**d)


class Config:
    def __init__(self, **kwargs):
        self.rvc = RVC(self)
        self.rvh = RVH(self)
        self.rvi = RVI(self)
        self.rvp = RVP(self)
        self.rvt = RVT(self)
        self.ost = OST(self)
        self.identifier = None
        self.update(**kwargs)

    def update(self, key=None, value=None, **kwargs):
        def _update_single(key, value):
            updated = False
            for rv in [self.rvc, self.rvi, self.rvh, self.rvp, self.rvt, self.ost]:
                # Note that in certain cases we might need to update a key
                # for more than one rv object.
                if rv.update(key, value):
                    updated = True
            if not updated:
                raise AttributeError(
                    f"No field named `{key}` found in any RV* conf class"
                )

        identifier = kwargs.pop("identifier", None)
        if identifier:
            self.identifier = identifier

        if key is None and value is None:
            for k, v in kwargs.items():
                _update_single(k, v)
        else:
            _update_single(key, value)

    def set_rv_file(self, fn):
        fn = Path(fn)
        if fn.name == "OstRandomNumbers.txt":
            self.ost.random_numbers_path = fn
        else:
            rvx = fn.suffixes[0][1:]  # get first suffix: eg.g. .rvt[.tpl]
            rvo = getattr(self, rvx, None) or self.ost
            rvo.content = fn.read_text()
            rvo.is_ostrich_tmpl = fn.suffixes[-1] == ".tpl"
            if not self.identifier:
                # Get the "true" stem if there are more than one suffixes
                identifier = fn
                while identifier.suffixes:
                    identifier = Path(identifier.stem)
                self.identifier = identifier.as_posix()
            # This is a sorry hack: I want to have rvi.run_name have a default of "run"
            # because I don't want to burden the user with setting it.. but the problem
            # is that externally supplied rv files might not have it (to be discussed)
            self.rvi.run_name = None



import datetime as dt
from dataclasses import replace

import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr

from ravenpy.config.commands import BasinIndexCommand, HRUState
from ravenpy.config.rvs import RVC
from ravenpy.models import GR4JCN
from ravenpy.utilities.data_assimilation import assimilate, perturbation
from ravenpy.utilities.testdata import get_local_testdata


def test_perturbation():
    ts = get_local_testdata(
        "raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
    )
    ds = xr.open_dataset(ts)

    tmax = ds.tmax.isel(time=slice(0, 10))
    p_tmax = perturbation(tmax, "norm", 0.01, members=50)
    np.testing.assert_allclose(p_tmax.mean("members"), tmax, rtol=0.1)

    rain = ds.rain.isel(time=slice(30, 60))
    p_rain = perturbation(rain, "gamma", 0.01, members=50)
    np.testing.assert_allclose(p_rain.mean("members"), rain, rtol=0.1)

    assert p_tmax.attrs == ds.tmax.attrs
    assert p_rain.attrs == ds.rain.attrs


class TestAssimilationGR4JCN:
    def test_simple(self):

        # get timeseries
        ts = get_local_testdata(
            "raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
        )

        # set number of members. Using 7 here to make it easier to find and debug.
        n_members = 7

        # Perturbation parameters for the assimilation, keyed by standard_name
        std = {
            "rainfall": 0.30,
            "prsn": 0.30,
            "tasmin": 2.0,
            "tasmax": 2.0,
            "water_volume_transport_in_river_channel": 0.15,
        }

        # Use the same random seed for both tasmin and tasmax
        rs = np.random.SeedSequence(None).generate_state(1)[0]
        seed = {"tasmin": rs, "tasmax": rs}

        # Perturbation distribution
        dists = {
            "pr": "gamma",
            "rainfall": "gamma",
            "prsn": "gamma",
            "water_volume_transport_in_river_channel": "rnorm",
        }

        qkey = "water_volume_transport_in_river_channel"
        if qkey not in std:
            raise ValueError("Assimilation requires perturbing the flow variable.")

        # Assimilation variables (from HRUStateVariable)
        assim_var = ("soil0", "soil1")

        # Assimilation periods
        assim_days = [10] + 4 * [3]

        # GR4JCN model instance
        model = GR4JCN()

        # set the start and end dates for the first assimilation period, warm-up
        start_date = dt.datetime(2000, 6, 1)
        end_date = start_date + dt.timedelta(days=sum(assim_days))

        # Set model options

        model.config.rvh.hrus = (
            GR4JCN.LandHRU(
                area=4250.6, elevation=843.0, latitude=54.4848, longitude=-123.3659
            ),
        )

        model.config.rvp.params = GR4JCN.Params(
            0.1353389, -0.005067198, 576.8007, 6.986121, 1.102917, 0.9224778
        )  # SALMON

        # ==== Initialization (just to get reasonable states) ====
        # Set initialization run options
        model.config.rvi.run_name = "init"
        model.config.rvi.start_date = start_date
        # Richard: what is the end date policy for the init run ?
        model.config.rvi.end_date = start_date + dt.timedelta(days=assim_days[0])

        # Run the model
        model([ts])

        # Extract final model states
        hru_state, basin_state = model.get_final_state()
        xa = n_members * [getattr(hru_state, key) for key in assim_var]
        hru_states = n_members * [hru_state]
        basin_states = n_members * [basin_state]

        # === Create perturbed time series for full assimilation period ====
        perturbed = {}
        for key, s in std.items():
            nc = model.config.rvt._var_cmds[key]

            with xr.open_dataset(nc.file_name_nc) as ds:
                da = ds.get(nc.var_name_nc).sel(time=slice(start_date, end_date))

                perturbed[key] = perturbation(
                    da,
                    dists.get(key, "norm"),
                    std=s,
                    seed=seed.get(key, None),
                    member=n_members,
                )

                # Save flow for later
                if key == qkey:
                    q_obs = da

        # Write to disk
        p_fn = model.workdir / "perturbed_forcing.nc"
        perturbed = xr.Dataset(perturbed)
        perturbed.to_netcdf(p_fn, mode="w")

        # ==== Assimilation ====
        q_assim = []
        sd = start_date

        for i, ndays in enumerate(assim_days):

            dates = [sd + dt.timedelta(days=x) for x in range(ndays)]
            model.config.rvi.end_date = dates[-1]
            model.config.rvi.run_name = f"assim_{i}"

            # Perform the first assimilation step here
            [xa, model] = assimilate(
                model, p_fn, q_obs, assim_var, basin_states, hru_states, dates
            )

            # Save streamflow simulation
            q_assim.append(model.q_sim.isel(nbasins=0))

            # Update the start-time for the next loop
            sd += dt.timedelta(days=ndays)
            model.config.rvi.start_date = sd

            # Get new initial conditions and feed assimilated values
            hru_states, basin_states = model.get_final_state()
            hru_states = [
                replace(hru_states[i], **dict(zip(assim_var, xa[:, i])))
                for i in range(n_members)
            ]

        q_assim = xr.concat(q_assim, dim="time")

        # ==== Reference run ====
        model.config.rvi.run_name = "ref"
        model.config.rvi.start_date = start_date
        model.config.rvi.end_date = end_date

        model.config.rvc.hru_states = {}
        model.config.rvc.basin_states = {}
        model.config.rvc.soil0 = None
        model.config.rvc.soil1 = 15

        model([ts])

        # We can now plot everything!
        plt.plot(q_assim.T, "r", label="Assimilated")  # plot the assimilated flows
        plt.plot(q_obs.T, "b", label="Observed")  # plot the observed flows
        plt.plot(
            model.q_sim, "g", label="Simulated"
        )  # plot the open_loop (simulation with no assimilation)
        # plt.legend()
        # plt.show()

        assert q_assim.shape[0] == n_members
        assert q_assim.shape[1] == sum(assim_days)



import datetime as dt

from ravenpy.models import GR4JCN
from ravenpy.utilities import regionalization as reg
from ravenpy.utilities.testdata import get_local_testdata


def test_regionalization():
    ts = get_local_testdata(
        "raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc"
    )
    model = "GR4JCN"
    nash, params = reg.read_gauged_params(model)
    variables = ["latitude", "longitude", "area", "forest"]
    props = reg.read_gauged_properties(variables)
    ungauged_props = {
        "latitude": 40.4848,
        "longitude": -103.3659,
        "area": 4250.6,
        "forest": 0.4,
    }

    hrus = (
        GR4JCN.LandHRU(
            area=4250.6, elevation=843.0, latitude=40.4848, longitude=-103.3659
        ),
    )

    qsim, ens = reg.regionalize(
        "SP_IDW",
        model,
        nash,
        params,
        props,
        ungauged_props,
        start_date=dt.datetime(2000, 1, 1),
        end_date=dt.datetime(2002, 1, 1),
        hrus=hrus,
        longitude=-103.3659,
        min_NSE=0.6,
        size=2,
        ts=ts,
    )

    assert qsim.max() > 1
    assert len(ens) == 2
    assert "realization" in ens.dims
    assert "param" in ens.dims



import logging
import math
import tempfile
from pathlib import Path
from typing import List, Optional, Union

import numpy as np

from . import gis_import_error_message

try:
    import rasterio
    from osgeo.gdal import Dataset, DEMProcessing
    from shapely.geometry import GeometryCollection, MultiPolygon, Polygon, shape
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

from ravenpy.utilities.geo import generic_raster_clip

# See: https://kokoalberti.com/articles/geotiff-compression-optimization-guide/
# or 'compress=deflate' or 'compress=zstd' or 'compress=lerc' or others
GDAL_TIFF_COMPRESSION_OPTION = "compress=lzw"

LOGGER = logging.getLogger("RavenPy")


def geom_prop(geom: Union[Polygon, MultiPolygon, GeometryCollection]) -> dict:
    """Return a dictionary of geometry properties.

    Parameters
    ----------
    geom : Union[Polygon, MultiPolygon, GeometryCollection]
      Geometry to analyze.

    Returns
    -------
    dict
      Dictionary storing polygon area, centroid location, perimeter and gravelius shape index.

    Notes
    -----
    Some of the properties should be computed using an equal-area projection.
    """

    geom = shape(geom)
    lon, lat = geom.centroid.x, geom.centroid.y
    if (lon > 180) or (lon < -180) or (lat > 90) or (lat < -90):
        LOGGER.warning("Shape centroid is not in decimal degrees.")
    area = geom.area
    length = geom.length
    gravelius = length / 2 / math.sqrt(math.pi * area)
    parameters = {
        "area": area,
        "centroid": (lon, lat),
        "perimeter": length,
        "gravelius": gravelius,
    }
    return parameters


def dem_prop(
    dem: Union[str, Path],
    geom: Union[Polygon, MultiPolygon, List[Union[Polygon, MultiPolygon]]] = None,
    directory: Union[str, Path] = None,
) -> dict:
    """Return raster properties for each geometry.

    This

    Parameters
    ----------
    dem : Union[str, Path]
      DEM raster in reprojected coordinates.
    geom : Union[Polygon, MultiPolygon, List[Union[Polygon, MultiPolygon]]]
      Geometry over which aggregate properties will be computed. If None compute properties over entire raster.
    directory : Union[str, Path]
      Folder to save the GDAL terrain analysis outputs.

    Returns
    -------
    dict
      Dictionary storing mean elevation [m], slope [deg] and aspect [deg].
    """

    fns = dict()
    fns["dem"] = (
        tempfile.NamedTemporaryFile(
            prefix="dem", suffix=".tiff", dir=directory, delete=False
        ).name
        if geom is not None
        else dem
    )
    for key in ["slope", "aspect"]:
        fns[key] = tempfile.NamedTemporaryFile(
            prefix=key, suffix=".tiff", dir=directory, delete=False
        ).name

    # Clip to relevant area or read original raster
    if geom is None:
        with rasterio.open(dem) as f:
            elevation = f.read(1, masked=True)
    else:
        generic_raster_clip(raster=dem, output=fns["dem"], geometry=geom)
        with rasterio.open(fns["dem"]) as f:
            elevation = f.read(1, masked=True)

    # Compute slope
    slope = gdal_slope_analysis(fns["dem"], set_output=fns["slope"])

    # Compute aspect
    aspect = gdal_aspect_analysis(fns["dem"], set_output=fns["aspect"])
    aspect_mean = circular_mean_aspect(aspect)

    return {"elevation": elevation.mean(), "slope": slope.mean(), "aspect": aspect_mean}


def gdal_slope_analysis(
    dem: Union[str, Path],
    set_output: Optional[Union[str, Path]] = None,
    units: str = "degree",
) -> np.ndarray:
    """Return the slope of the terrain from the DEM.

    The slope is the magnitude of the gradient of the elevation.

    Parameters
    ----------
    dem : Union[str, Path]
      Path to file storing DEM.
    set_output : Union[str, Path]
      If set to a valid filepath, will write to this path, otherwise will use an in-memory gdal.Dataset.
    units : str
      Slope units. Default: 'degree'.

    Returns
    -------
    np.ndarray
      Slope array.

    Notes
    -----
    Ensure that the DEM is in a *projected coordinate*, not a geographic coordinate system, so that the
    horizontal scale is the same as the vertical scale (m).

    """
    if isinstance(dem, Path):
        dem = str(dem)
    if set_output:
        if isinstance(set_output, (str, Path)):
            set_output = str(set_output)
            DEMProcessing(
                set_output,
                dem,
                "slope",
                slopeFormat=units,
                format="GTiff",
                band=1,
                creationOptions=[GDAL_TIFF_COMPRESSION_OPTION],
            )
            with rasterio.open(set_output) as src:
                return np.ma.masked_values(src.read(1), value=-9999)
        else:
            raise ValueError()
    else:
        set_output = DEMProcessing(
            "",
            dem,
            "slope",
            slopeFormat=units,
            format="MEM",
            band=1,
        )
        return np.ma.masked_values(set_output.ReadAsArray(), value=-9999)


def gdal_aspect_analysis(
    dem: Union[str, Path],
    set_output: Union[str, Path, bool] = False,
    flat_values_are_zero: bool = False,
) -> Union[np.ndarray, Dataset]:
    """Return the aspect of the terrain from the DEM.

    The aspect is the compass direction of the steepest slope (0: North, 90: East, 180: South, 270: West).

    Parameters
    ----------
    dem : Union[str, Path]
      Path to file storing DEM.
    set_output : Union[str, Path, bool]
      If set to a valid filepath, will write to this path, otherwise will use an in-memory gdal.Dataset.
    flat_values_are_zero: bool
      Designate flat values with value zero. Default: -9999.

    Returns
    -------
    np.ndarray
      Aspect array.

    Notes
    -----
    Ensure that the DEM is in a *projected coordinate*, not a geographic coordinate system, so that the
    horizontal scale is the same as the vertical scale (m).
    """
    if isinstance(dem, Path):
        dem = str(dem)
    if set_output:
        if isinstance(set_output, (str, Path)):
            set_output = str(set_output)
            DEMProcessing(
                destName=set_output,
                srcDS=dem,
                processing="aspect",
                zeroForFlat=flat_values_are_zero,
                format="GTiff",
                band=1,
                creationOptions=[GDAL_TIFF_COMPRESSION_OPTION],
            )
            with rasterio.open(set_output) as src:
                return np.ma.masked_values(src.read(1), value=-9999)
        else:
            raise ValueError()

    else:
        set_output = DEMProcessing(
            destName="",
            srcDS=dem,
            processing="aspect",
            zeroForFlat=flat_values_are_zero,
            format="MEM",
            band=1,
        )
        return np.ma.masked_values(set_output.ReadAsArray(), value=-9999)


def circular_mean_aspect(angles: np.ndarray) -> np.ndarray:
    """Return the mean angular aspect based on circular arithmetic approach

    Parameters
    ----------
    angles: np.ndarray
      Array of aspect angles

    Returns
    -------
    np.ndarray
      Circular mean of aspect array.
    """
    # Circular statistics needed for mean angular aspect
    # Example from: https://gis.stackexchange.com/a/147135/65343

    n = len(angles)
    sine_mean = np.divide(np.sum(np.sin(np.radians(np.ma.masked_array(angles)))), n)
    cosine_mean = np.divide(np.sum(np.cos(np.radians(np.ma.masked_array(angles)))), n)
    vector_mean = np.arctan2(sine_mean, cosine_mean)
    degrees = np.degrees(vector_mean)

    if degrees < 0:
        return degrees + 360
    return degrees
